from PIL import Image
import os
import jax
import jax.numpy as jnp

from jaxmarl import make
from jaxmarl.environments.storm.storm_env import Items

action=1
render_agent_view = False
num_outer_steps = 3
# num_inner_steps = 68
#num_agents=8
num_agents=2
num_inner_steps=152

rng = jax.random.PRNGKey(18)
env = make('storm', 
        num_inner_steps=num_inner_steps, 
        num_outer_steps=num_outer_steps, 
        num_agents=num_agents, 
        fixed_coin_location=True,
        payoff_matrix=jnp.array([[[3, 0], [5, 1]], [[3, 5], [0, 1]]]),
        freeze_penalty=5,)
num_actions = env.action_space(0).n
print('NUM ACTIONS: ', num_actions)
obs, old_state = env.reset(rng)


def pos_remove(i, val):
        grid, old_pos = val
        grid = grid.at[
            (old_pos[i, 0], old_pos[i,1])
        ].set(jnp.int8(Items.empty))
        return (grid, old_pos)

def pos_add(i, val):
        grid, new_pos = val
        grid = grid.at[(new_pos[i,0], new_pos[i,1])].set(
            jnp.int8(i+1)
        )
        return (grid, new_pos)  


# grid, _ = jax.lax.fori_loop(0, num_agents, pos_remove, (old_state.grid, old_state.agent_positions))
# new_agent_pos = jnp.array([[7, 0, 0],
#     [7, 1, 2],
#     [7, 2, 2]], dtype=jnp.int8)
# grid, _ = jax.lax.fori_loop(0, num_agents, pos_add, (grid, new_agent_pos))
# old_state= old_state.replace(agent_positions=new_agent_pos)
# old_state= old_state.replace(agent_inventories=jnp.array([[1., 1.],
#     [1., 1.],
#     [1., 1.]]))
# old_state= old_state.replace(grid=grid)

pics = []
pics1 = []
pics2 = []

# new_pos = []

img = env.render(old_state)
if not os.path.exists('state_pics'):
    os.makedirs('state_pics')
Image.fromarray(img).save("state_pics/init_state.png")
# img1 = env.render_agent_view(old_state, agent=0)
# img2 = env.render_agent_view(old_state, agent=1)
pics.append(img)

int_action = {
    0: "left",
    1: "right",
    2: "forward",
    3: "interact",
    4: "stay",
}

key_int = {"w": 2, "a": 0, "s": 4, "d": 1, " ": 4}
env.step_env = jax.jit(env.step_env)

for t in range(num_outer_steps * num_inner_steps):
    rng, *rngs = jax.random.split(rng, num_agents+1)
    # a1 = jnp.array(2)
    # a2 = jnp.array(4)
    actions = [jax.random.choice(
        rngs[a], a=num_actions, p=jnp.array([0.1, 0.1,0.5, 0.1, 0.2])
    ) for a in range(num_agents)]
    # actions = [jax.random.choice(
    #     rngs[a], a=num_actions, p=jnp.array([0.0, 0.0,1.0, 0.0, 0.0])
    # ) for a in range(num_agents)]
    print('###################')
    print(f'timestep: {t} to {t+1}')
    print([action.item() for action in actions], 'actions')
    print("###################")

    obs, state, reward, done, info = env.step_env(
        rng, old_state, [a*action for a in actions]
    )
    print('outer t', state.outer_t)
    print('inner t', state.inner_t)
    print('done', done)

    img = env.render(state)
    Image.fromarray(img).save(f"state_pics/state_{t+1}.png")
    pics.append(img)

    if render_agent_view:
        img1 = env.render_agent_view(state, agent=0)
        img2 = env.render_agent_view(state, agent=1)
        img3 = env.render_agent_view(state, agent=2)
        Image.fromarray(img1).save(f"state_pics/view_ag1_{t+1}.png")
        Image.fromarray(img2).save(f"state_pics/view_ag2_{t+1}.png")
        Image.fromarray(img3).save(f"state_pics/view_ag3_{t+1}.png")
        pics1.append(img1)
        pics2.append(img2)
    old_state = state
    # if True:
    # if  t==74 or t==75 or t==76:
        # jax.debug.breakpoint()
    # if t>76:
    #     raise
print("Saving GIF")
pics = [Image.fromarray(img) for img in pics]        
pics[0].save(
    "state.gif",
    format="GIF",
    save_all=True,
    optimize=False,
    append_images=pics[1:],
    duration=1000,
    loop=0,
)

if render_agent_view:
    pics1 = [Image.fromarray(img) for img in pics1]
    pics2 = [Image.fromarray(img) for img in pics2]
    pics1[0].save(
        "agent1.gif",
        format="GIF",
        save_all=True,
        optimize=False,
        append_images=pics1[1:],
        duration=1000,
        loop=0,
    )
    pics2[0].save(
        "agent2.gif",
        format="GIF",
        save_all=True,
        optimize=False,
        append_images=pics2[1:],
        duration=1000,
        loop=0,
    )


import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from jaxmarl.environments.mpe.simple import SimpleMPE, State
from jaxmarl.environments.mpe.default_params import *
from jaxmarl.environments.spaces import Box, Discrete

SPEAKER = "speaker_0"
LISTENER = "listener_0"
AGENT_NAMES = [SPEAKER, LISTENER]

OBS_COLOURS = [(166, 38, 38), (38, 166, 38), (38, 38, 166)]

class SimpleSpeakerListenerMPE(SimpleMPE):
    def __init__(
        self,
        num_agents=2,
        num_landmarks=3,
        action_type=DISCRETE_ACT,
        **kwargs,
    ):
        assert num_agents == 2, "SimpleSpeakerListnerMPE only supports 2 agents"
        assert num_landmarks == 3, "SimpleSpeakerListnerMPE only supports 3 landmarks"

        num_entities = num_agents + num_landmarks

        dim_c = 3
        # collaborative bool .. ?

        agents = AGENT_NAMES

        landmarks = ["landmark {}".format(i) for i in range(num_landmarks)]

        # Action and observation spaces

        if action_type == DISCRETE_ACT:
            action_spaces = {
                SPEAKER: Discrete(3),
                LISTENER: Discrete(5),
            }
        elif action_type == CONTINUOUS_ACT:
            action_spaces = {
                SPEAKER: Box(0.0, 1.0, (3,)),
                LISTENER: Box(0.0, 1.0, (5,)),
            }
        else:
            raise NotImplementedError("Action type not implemented")

        observation_spaces = {
            SPEAKER: Box(-jnp.inf, jnp.inf, (3,)),
            LISTENER: Box(-jnp.inf, jnp.inf, (11,)),
        }

        colour = (
            [ADVERSARY_COLOUR]
            + [AGENT_COLOUR]
            + OBS_COLOURS
        )

        # Parameters
        rad = jnp.concatenate(
            [jnp.full((num_agents), 0.075), jnp.full((num_landmarks), 0.04)]
        )
        moveable = jnp.concatenate(
            [jnp.array([False]), jnp.array([True]), jnp.full((num_landmarks), False)]
        )
        silent = jnp.array([0, 1])
        collide = jnp.full((num_entities), False)

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            action_spaces=action_spaces,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            moveable=moveable,
            silent=silent,
            collide=collide,
            **kwargs,
        )

    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        key_a, key_l, key_g = jax.random.split(key, 3)

        p_pos = jnp.concatenate(
            [
                jax.random.uniform(key_a, (self.num_agents, 2), minval=-1, maxval=+1),
                jax.random.uniform(
                    key_l, (self.num_landmarks, 2), minval=-1.0, maxval=+1.0
                ),
            ]
        )

        g_idx = jax.random.randint(key_g, (), minval=0, maxval=self.num_landmarks)

        state = State(
            p_pos=p_pos,
            p_vel=jnp.zeros((self.num_entities, self.dim_p)),
            c=jnp.zeros((self.num_agents, self.dim_c)),
            done=jnp.full((self.num_agents), False),
            step=0,
            goal=g_idx,
        )

        return self.get_obs(state), state

    def set_actions(self, actions: Dict):
        """Extract u and c actions for all agents from actions Dict."""

        """actions = jnp.array([actions[i] for i in self.agents]).reshape((self.num_agents, -1))"""

        return self.action_decoder(None, actions)

    def _decode_continuous_action(
        self,
        a_idx: int,
        action: chex.Array,
    ) -> Tuple[chex.Array, chex.Array]:
        u = jnp.zeros((self.num_agents, self.dim_p))
        c = jnp.zeros((self.num_agents, self.dim_c))
        c = c.at[0].set(action[SPEAKER])

        u_act = action[LISTENER]

        u_act = jnp.array([u_act[2] - u_act[1], u_act[4] - u_act[3]]) * self.accel[1]
        u = u.at[1].set(u_act)

        return u, c

    def _decode_discrete_action(
        self,
        a_idx: int,
        action: chex.Array,
    ) -> Tuple[chex.Array, chex.Array]:
        u = jnp.zeros((self.num_agents, self.dim_p))
        c = jnp.zeros((self.num_agents, self.dim_c))

        c = c.at[0, action[SPEAKER]].set(1.0)

        idx = jax.lax.select(action[LISTENER] <= 2, 0, 1)
        u_val = jax.lax.select(action[LISTENER] % 2 == 0, 1.0, -1.0) * (
            action[LISTENER] != 0
        )
        u = u.at[1, idx].set(u_val)
        u = u * self.accel[1] * self.moveable[1]
        return u, c

    def rewards(
        self,
        state: State,
    ) -> Dict[str, float]:
        r = -1 * jnp.sum(
            jnp.square(state.p_pos[1] - state.p_pos[state.goal + self.num_agents])
        )
        return {a: r for a in self.agents}

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        goal_colour = jnp.full((3,), 0.15)
        goal_colour = goal_colour.at[state.goal].set(0.65)

        dist = state.p_pos[self.num_agents :] - state.p_pos[1]
        comm = state.c[0]

        def _speaker():
            return goal_colour

        def _listener():
            return jnp.concatenate([state.p_vel[1], dist.flatten(), comm])

        return {SPEAKER: _speaker(), LISTENER: _listener()}



import jax
import jax.numpy as jnp
import chex
from typing import Tuple, Dict
from functools import partial
from jaxmarl.environments.mpe.simple import SimpleMPE, State
from jaxmarl.environments.mpe.default_params import *
from jaxmarl.environments.spaces import Box


class SimpleSpreadMPE(SimpleMPE):
    def __init__(
        self,
        num_agents=3,
        num_landmarks=3,
        local_ratio=0.5,
        action_type=DISCRETE_ACT,
        **kwargs,
    ):
        dim_c = 2  # NOTE follows code rather than docs

        # Action and observation spaces
        agents = ["agent_{}".format(i) for i in range(num_agents)]
        landmarks = ["landmark {}".format(i) for i in range(num_landmarks)]

        observation_spaces = {
            i:Box(-jnp.inf, jnp.inf, (4+(num_agents-1)*4+(num_landmarks*2),)) 
            for i in agents
        }

        colour = [AGENT_COLOUR] * num_agents + [OBS_COLOUR] * num_landmarks

        # Env specific parameters
        self.local_ratio = local_ratio
        assert (
            self.local_ratio >= 0.0 and self.local_ratio <= 1.0
        ), "local_ratio must be between 0.0 and 1.0"

        # Parameters
        rad = jnp.concatenate(
            [jnp.full((num_agents), 0.15), jnp.full((num_landmarks), 0.05)]
        )
        collide = jnp.concatenate(
            [jnp.full((num_agents), True), jnp.full((num_landmarks), False)]
        )

        super().__init__(
            num_agents=num_agents,
            agents=agents,
            num_landmarks=num_landmarks,
            landmarks=landmarks,
            action_type=action_type,
            observation_spaces=observation_spaces,
            dim_c=dim_c,
            colour=colour,
            rad=rad,
            collide=collide,
            **kwargs,
        )

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        @partial(jax.vmap, in_axes=(0))
        def _common_stats(aidx: int):
            """Values needed in all observations"""

            landmark_pos = (
                state.p_pos[self.num_agents :] - state.p_pos[aidx]
            )  # Landmark positions in agent reference frame

            # Zero out unseen agents with other_mask
            other_pos = state.p_pos[: self.num_agents] - state.p_pos[aidx]

            # use jnp.roll to remove ego agent from other_pos and other_vel arrays
            other_pos = jnp.roll(other_pos, shift=self.num_agents - aidx - 1, axis=0)[
                : self.num_agents - 1
            ]
            comm = jnp.roll(
                state.c[: self.num_agents], shift=self.num_agents - aidx - 1, axis=0
            )[: self.num_agents - 1]

            other_pos = jnp.roll(other_pos, shift=aidx, axis=0)
            comm = jnp.roll(comm, shift=aidx, axis=0)

            return landmark_pos, other_pos, comm

        landmark_pos, other_pos, comm = _common_stats(self.agent_range)

        def _obs(aidx: int):
            return jnp.concatenate(
                [
                    state.p_vel[aidx].flatten(),  # 2
                    state.p_pos[aidx].flatten(),  # 2
                    landmark_pos[aidx].flatten(),  # 5, 2
                    other_pos[aidx].flatten(),  # 5, 2
                    comm[aidx].flatten(),
                ]
            )

        obs = {a: _obs(i) for i, a in enumerate(self.agents)}
        return obs

    def rewards(self, state: State) -> Dict[str, float]:
        @partial(jax.vmap, in_axes=(0, None))
        def _collisions(agent_idx: int, other_idx: int):
            return jax.vmap(self.is_collision, in_axes=(None, 0, None))(
                agent_idx,
                other_idx,
                state,
            )

        c = _collisions(
            self.agent_range,
            self.agent_range,
        )  # [agent, agent, collison]

        def _agent_rew(aidx: int, collisions: chex.Array):
            rew = -1 * jnp.sum(collisions[aidx])
            return rew

        def _land(land_pos: chex.Array):
            d = state.p_pos[: self.num_agents] - land_pos
            return -1 * jnp.min(jnp.linalg.norm(d, axis=1))

        global_rew = jnp.sum(jax.vmap(_land)(state.p_pos[self.num_agents :]))

        rew = {
            a: _agent_rew(i, c) * self.local_ratio + global_rew * (1 - self.local_ratio)
            for i, a in enumerate(self.agents)
        }
        return rew



import jax.numpy as jnp
import jax


def downsample(img, factor):
    """
    Downsample an image along both dimensions by some factor
    """
    assert img.shape[0] % factor == 0
    assert img.shape[1] % factor == 0

    img = img.reshape(
        [img.shape[0] // factor, factor, img.shape[1] // factor, factor, 3]
    )
    img = img.mean(axis=3)
    img = img.mean(axis=1)

    # convert back to uint8
    img = img.astype(jnp.uint8)

    return img


def fill_coords(img, fn, color):
    """
    Fill pixels of an image with coordinates matching a filter function
    """

    def _mask_fn(y, x):
        yf = (y + 0.5) / img.shape[0]
        xf = (x + 0.5) / img.shape[1]
        return fn(xf, yf)

    ys, xs = jnp.indices(img.shape[:2])
    mask = jax.vmap(jax.vmap(_mask_fn))(ys, xs)

    color_img = jnp.full_like(img, color)
    return jnp.where(mask[:, :, None], color_img, img)


def rotate_fn(fin, cx, cy, theta):
    def fout(x, y):
        x = x - cx
        y = y - cy

        x2 = cx + x * jnp.cos(-theta) - y * jnp.sin(-theta)
        y2 = cy + y * jnp.cos(-theta) + x * jnp.sin(-theta)

        return fin(x2, y2)

    return fout


def point_in_line(x0, y0, x1, y1, r):
    p0 = jnp.array([x0, y0])
    p1 = jnp.array([x1, y1])
    dir = p1 - p0
    dist = jnp.linalg.norm(dir)
    dir = dir / dist

    xmin = min(x0, x1) - r
    xmax = max(x0, x1) + r
    ymin = min(y0, y1) - r
    ymax = max(y0, y1) + r

    def fn(x, y):
        # Fast, early escape test
        if x < xmin or x > xmax or y < ymin or y > ymax:
            return False

        q = jnp.array([x, y])
        pq = q - p0

        # Closest point on line
        a = jnp.dot(pq, dir)
        a = jnp.clip(a, 0, dist)
        p = p0 + a * dir

        dist_to_line = jnp.linalg.norm(q - p)
        return dist_to_line <= r

    return fn


def point_in_circle(cx, cy, r):
    def fn(x, y):
        return (x - cx) * (x - cx) + (y - cy) * (y - cy) <= r * r

    return fn


def point_in_rect(xmin, xmax, ymin, ymax):
    def fn(x, y):
        return (x >= xmin) & (x <= xmax) & (y >= ymin) & (y <= ymax)

    return fn


def point_in_triangle(a, b, c):
    a = jnp.array(a)
    b = jnp.array(b)
    c = jnp.array(c)

    def fn(x, y):
        v0 = c - a
        v1 = b - a
        v2 = jnp.array((x, y)) - a

        # Compute dot products
        dot00 = jnp.dot(v0, v0)
        dot01 = jnp.dot(v0, v1)
        dot02 = jnp.dot(v0, v2)
        dot11 = jnp.dot(v1, v1)
        dot12 = jnp.dot(v1, v2)

        # Compute barycentric coordinates
        inv_denom = 1 / (dot00 * dot11 - dot01 * dot01)
        u = (dot11 * dot02 - dot01 * dot12) * inv_denom
        v = (dot00 * dot12 - dot01 * dot02) * inv_denom

        # Check if point is in triangle
        return (u >= 0) & (v >= 0) & (u + v < 1)

    return fn


def highlight_img(img, color=(255, 255, 255), alpha=0.30):
    """
    Add highlighting to an image
    """
    blend_img = img + alpha * (jnp.array(color, dtype=jnp.uint8) - img)
    res_img = jnp.clip(blend_img, 0, 255).astype(jnp.uint8)

    return res_img



import math
from jaxmarl.environments.overcooked_v2.utils import compute_view_box
from jaxmarl.viz.window import Window
import jaxmarl.viz.grid_rendering_v2 as rendering
import jax
import jax.numpy as jnp
from jaxmarl.environments.overcooked_v2.common import StaticObject, DynamicObject
from jaxmarl.environments.overcooked_v2.settings import (
    POT_COOK_TIME,
    INDICATOR_ACTIVATION_TIME,
)
import imageio
from functools import partial

TILE_PIXELS = 32

COLORS = {
    "red": jnp.array([255, 0, 0], dtype=jnp.uint8),
    "green": jnp.array([0, 255, 0], dtype=jnp.uint8),
    "blue": jnp.array([0, 0, 255], dtype=jnp.uint8),
    "purple": jnp.array([160, 32, 240], dtype=jnp.uint8),
    "yellow": jnp.array([255, 255, 0], dtype=jnp.uint8),
    "grey": jnp.array([100, 100, 100], dtype=jnp.uint8),
    "white": jnp.array([255, 255, 255], dtype=jnp.uint8),
    "black": jnp.array([25, 25, 25], dtype=jnp.uint8),
    "orange": jnp.array([230, 180, 0], dtype=jnp.uint8),
    "pink": jnp.array([255, 105, 180], dtype=jnp.uint8),
    "brown": jnp.array([139, 69, 19], dtype=jnp.uint8),
    "cyan": jnp.array([0, 255, 255], dtype=jnp.uint8),
    "light_blue": jnp.array([173, 216, 230], dtype=jnp.uint8),
    "dark_green": jnp.array([0, 150, 0], dtype=jnp.uint8),
}

INGREDIENT_COLORS = jnp.array(
    [
        COLORS["yellow"],
        COLORS["dark_green"],
        COLORS["purple"],
        COLORS["cyan"],
        COLORS["red"],
        COLORS["orange"],
        COLORS["purple"],
        COLORS["blue"],
        COLORS["pink"],
        COLORS["brown"],
    ]
)


AGENT_COLORS = jnp.array(
    [
        COLORS["red"],
        COLORS["blue"],
        COLORS["green"],
        COLORS["purple"],
        COLORS["yellow"],
        COLORS["orange"],
    ]
)


class OvercookedV2Visualizer:
    """
    Manages a window and renders contents of EnvState instances to it.
    """

    tile_cache = {}

    def __init__(self, tile_size=TILE_PIXELS, subdivs=3):
        self.window = None

        self.tile_size = tile_size
        self.subdivs = subdivs

    def _lazy_init_window(self):
        if self.window is None:
            self.window = Window("Overcooked V2")

    def show(self, block=False):
        self._lazy_init_window()
        self.window.show(block=block)

    def render(self, state, agent_view_size=None):
        """Method for rendering the state in a window. Esp. useful for interactive mode."""
        self._lazy_init_window()

        img = self._render_state(state, agent_view_size)

        self.window.show_img(img)

    def animate(self, state_seq, filename="animation.gif", agent_view_size=None):
        """Animate a gif give a state sequence and save if to file."""

        frame_seq = jax.vmap(self._render_state, in_axes=(0, None))(
            state_seq, agent_view_size
        )
        # print("frame_seq", frame_seq)
        # print("frame_seq.shape", frame_seq.shape)
        # print("frame_seq.dtype", frame_seq.dtype)

        imageio.mimsave(filename, frame_seq, "GIF", duration=0.5)

    def render_sequence(self, state_seq, agent_view_size=None):
        frame_seq = jax.vmap(self._render_state, in_axes=(0, None))(
            state_seq, agent_view_size
        )
        return frame_seq

    @classmethod
    def _encode_agent_extras(cls, direction, idx):
        return direction | (idx << 2)

    @classmethod
    def _decode_agent_extras(cls, extras):
        direction = extras & 0x3
        idx = extras >> 2
        return direction, idx

    @partial(jax.jit, static_argnums=(0, 2))
    def _render_state(self, state, agent_view_size=None):
        """
        Render the state
        """

        grid = state.grid
        agents = state.agents
        recipe = state.recipe

        num_agents = agents.dir.shape[0]

        def _include_agents(grid, x):
            agent, idx = x
            pos = agent.pos
            inventory = agent.inventory
            direction = agent.dir

            # we have to do the encoding because we don't really have a way to also pass the agent's id
            extra_info = OvercookedV2Visualizer._encode_agent_extras(direction, idx)

            new_grid = grid.at[pos.y, pos.x].set(
                [StaticObject.AGENT, inventory, extra_info]
            )
            return new_grid, None

        grid, _ = jax.lax.scan(_include_agents, grid, (agents, jnp.arange(num_agents)))

        static_objects = grid[:, :, 0]
        ingredients = grid[:, :, 1]
        extra_info = grid[:, :, 2]

        recipe_indicator_mask = static_objects == StaticObject.RECIPE_INDICATOR
        button_recipe_indicator_mask = (
            static_objects == StaticObject.BUTTON_RECIPE_INDICATOR
        ) & (extra_info > 0)

        new_ingredients_layer = jnp.where(
            recipe_indicator_mask | button_recipe_indicator_mask,
            recipe | DynamicObject.COOKED | DynamicObject.PLATE,
            ingredients,
        )
        grid = grid.at[:, :, 1].set(new_ingredients_layer)

        highlight_mask = jnp.zeros(grid.shape[:2], dtype=bool)
        if agent_view_size:
            for x, y in zip(agents.pos.x, agents.pos.y):
                x_low, x_high, y_low, y_high = compute_view_box(
                    x, y, agent_view_size, grid.shape[0], grid.shape[1]
                )

                row_mask = jnp.arange(grid.shape[0])
                col_mask = jnp.arange(grid.shape[1])

                row_mask = (row_mask >= y_low) & (row_mask < y_high)
                col_mask = (col_mask >= x_low) & (col_mask < x_high)

                agent_mask = row_mask[:, None] & col_mask[None, :]

                highlight_mask |= agent_mask

        # Render the whole grid
        img = self._render_grid(grid, highlight_mask)
        return img

    @staticmethod
    def _render_dynamic_item(
        ingredients,
        img,
        plate_fn=rendering.point_in_circle(0.5, 0.5, 0.3),
        ingredient_fn=rendering.point_in_circle(0.5, 0.5, 0.15),
        dish_positions=jnp.array([(0.5, 0.4), (0.4, 0.6), (0.6, 0.6)]),
    ):
        def _no_op(img, ingredients):
            return img

        def _render_plate(img, ingredients):
            return rendering.fill_coords(img, plate_fn, COLORS["white"])

        def _render_ingredient(img, ingredients):
            idx = DynamicObject.get_ingredient_idx(ingredients)
            return rendering.fill_coords(img, ingredient_fn, INGREDIENT_COLORS[idx])

        def _render_dish(img, ingredients):
            img = rendering.fill_coords(img, plate_fn, COLORS["white"])
            ingredient_indices = DynamicObject.get_ingredient_idx_list_jit(ingredients)

            for idx, ingredient_idx in enumerate(ingredient_indices):
                color = INGREDIENT_COLORS[ingredient_idx]
                pos = dish_positions[idx]
                ingredient_fn = rendering.point_in_circle(pos[0], pos[1], 0.1)
                img_ing = rendering.fill_coords(img, ingredient_fn, color)

                img = jax.lax.select(ingredient_idx != -1, img_ing, img)

            return img

        branches = jnp.array(
            [
                ingredients == 0,
                ingredients == DynamicObject.PLATE,
                DynamicObject.is_ingredient(ingredients),
                ingredients & DynamicObject.COOKED,
            ]
        )
        branch_idx = jnp.argmax(branches)

        img = jax.lax.switch(
            branch_idx,
            [_no_op, _render_plate, _render_ingredient, _render_dish],
            img,
            ingredients,
        )

        return img

    @staticmethod
    def _render_cell(cell, img):
        static_object = cell[0]

        def _render_empty(cell, img):
            return img

        def _render_wall(cell, img):
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            img = OvercookedV2Visualizer._render_dynamic_item(cell[1], img)

            return img

        def _render_agent(cell, img):
            tri_fn = rendering.point_in_triangle(
                (0.12, 0.19),
                (0.87, 0.50),
                (0.12, 0.81),
            )

            direction, idx = OvercookedV2Visualizer._decode_agent_extras(cell[2])

            # A bit hacky, but needed so that actions order matches the one of Overcooked-AI
            direction_reordering = jnp.array([3, 1, 0, 2])
            direction = direction_reordering[direction]

            agent_color = AGENT_COLORS[idx]

            tri_fn = rendering.rotate_fn(
                tri_fn, cx=0.5, cy=0.5, theta=0.5 * math.pi * direction
            )
            img = rendering.fill_coords(img, tri_fn, agent_color)

            img = OvercookedV2Visualizer._render_dynamic_item(
                cell[1],
                img,
                plate_fn=rendering.point_in_circle(0.75, 0.75, 0.2),
                ingredient_fn=rendering.point_in_circle(0.75, 0.75, 0.15),
                dish_positions=jnp.array([(0.65, 0.65), (0.85, 0.65), (0.75, 0.85)]),
            )

            return img

        def _render_agent_self(cell, img):
            # Note: This should not ever be called
            return img

        def _render_goal(cell, img):
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0.1, 0.9, 0.1, 0.9), COLORS["green"]
            )

            return img

        def _render_pot(cell, img):
            return OvercookedV2Visualizer._render_pot(cell, img)

        def _render_recipe_indicator(cell, img):
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0.1, 0.9, 0.1, 0.9), COLORS["brown"]
            )
            img = OvercookedV2Visualizer._render_dynamic_item(cell[1], img)

            return img

        def _render_button_recipe_indicator(cell, img):
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0.1, 0.9, 0.1, 0.9), COLORS["brown"]
            )
            img = OvercookedV2Visualizer._render_dynamic_item(cell[1], img)

            time_left = cell[2]
            progress_fn = rendering.point_in_rect(
                0.1,
                0.9 - (0.9 - 0.1) / INDICATOR_ACTIVATION_TIME * time_left,
                0.83,
                0.88,
            )
            img_timer = rendering.fill_coords(img, progress_fn, COLORS["green"])

            button_fn = rendering.point_in_circle(0.5, 0.5, 0.2)
            img_button = rendering.fill_coords(img, button_fn, COLORS["red"])

            img = jax.lax.select(time_left > 0, img_timer, img_button)
            return img

        def _render_plate_pile(cell, img):
            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            plate_fns = [
                rendering.point_in_circle(*coord, 0.2)
                for coord in [(0.3, 0.3), (0.75, 0.42), (0.4, 0.75)]
            ]
            for plate_fn in plate_fns:
                img = rendering.fill_coords(img, plate_fn, COLORS["white"])
            return img

        def _render_ingredient_pile(cell, img):
            ingredient_idx = cell[0] - StaticObject.INGREDIENT_PILE_BASE

            img = rendering.fill_coords(
                img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
            )
            ingredient_fns = [
                rendering.point_in_circle(*coord, 0.15)
                for coord in [
                    (0.5, 0.15),
                    (0.3, 0.4),
                    (0.8, 0.35),
                    (0.4, 0.8),
                    (0.75, 0.75),
                ]
            ]

            for ingredient_fn in ingredient_fns:
                img = rendering.fill_coords(
                    img, ingredient_fn, INGREDIENT_COLORS[ingredient_idx]
                )

            return img

        render_fns_dict = {
            StaticObject.EMPTY: _render_empty,
            StaticObject.WALL: _render_wall,
            StaticObject.AGENT: _render_agent,
            StaticObject.SELF_AGENT: _render_agent_self,
            StaticObject.GOAL: _render_goal,
            StaticObject.POT: _render_pot,
            StaticObject.RECIPE_INDICATOR: _render_recipe_indicator,
            StaticObject.BUTTON_RECIPE_INDICATOR: _render_button_recipe_indicator,
            StaticObject.PLATE_PILE: _render_plate_pile,
        }

        render_fns = [_render_empty] * (max(render_fns_dict.keys()) + 2)
        for key, value in render_fns_dict.items():
            render_fns[key] = value
        render_fns[-1] = _render_ingredient_pile

        branch_idx = jnp.clip(static_object, 0, len(render_fns) - 1)

        return jax.lax.switch(
            branch_idx,
            render_fns,
            cell,
            img,
        )

    @staticmethod
    def _render_pot(cell, img):
        ingredients = cell[1]
        time_left = cell[2]

        is_cooking = time_left > 0
        is_cooked = (ingredients & DynamicObject.COOKED) != 0
        is_idle = ~is_cooking & ~is_cooked
        ingredients = DynamicObject.get_ingredient_idx_list_jit(ingredients)
        has_ingredients = ingredients[0] != -1

        img = rendering.fill_coords(
            img, rendering.point_in_rect(0, 1, 0, 1), COLORS["grey"]
        )

        ingredient_fns = [
            rendering.point_in_circle(*coord, 0.13)
            for coord in [(0.23, 0.33), (0.77, 0.33), (0.50, 0.33)]
        ]

        for i, ingredient_idx in enumerate(ingredients):
            img_ing = rendering.fill_coords(
                img, ingredient_fns[i], INGREDIENT_COLORS[ingredient_idx]
            )
            img = jax.lax.select(ingredient_idx != -1, img_ing, img)

        pot_fn = rendering.point_in_rect(0.1, 0.9, 0.33, 0.9)
        lid_fn = rendering.point_in_rect(0.1, 0.9, 0.21, 0.25)
        handle_fn = rendering.point_in_rect(0.4, 0.6, 0.16, 0.21)

        lid_fn_open = rendering.rotate_fn(lid_fn, cx=0.1, cy=0.25, theta=-0.1 * math.pi)
        handle_fn_open = rendering.rotate_fn(
            handle_fn, cx=0.1, cy=0.25, theta=-0.1 * math.pi
        )
        pot_open = is_idle & has_ingredients

        img = rendering.fill_coords(img, pot_fn, COLORS["black"])

        img_closed = rendering.fill_coords(img, lid_fn, COLORS["black"])
        img_closed = rendering.fill_coords(img_closed, handle_fn, COLORS["black"])

        img_open = rendering.fill_coords(img, lid_fn_open, COLORS["black"])
        img_open = rendering.fill_coords(img_open, handle_fn_open, COLORS["black"])

        img = jax.lax.select(pot_open, img_open, img_closed)

        # Render progress bar
        progress_fn = rendering.point_in_rect(
            0.1, 0.9 - (0.9 - 0.1) / POT_COOK_TIME * time_left, 0.83, 0.88
        )
        img_timer = rendering.fill_coords(img, progress_fn, COLORS["green"])
        img = jax.lax.select(is_cooking, img_timer, img)

        return img

    def _render_tile(
        self,
        obj,
        highlight=False,
    ):
        """
        Render a tile and cache the result
        """
        # key = (*obj.tolist(), highlight, tile_size)

        # if key in OvercookedV2Visualizer.tile_cache:
        #     return OvercookedV2Visualizer.tile_cache[key]

        img = jnp.zeros(
            shape=(self.tile_size * self.subdivs, self.tile_size * self.subdivs, 3),
            dtype=jnp.uint8,
        )

        # Draw the grid lines (top and left edges)
        img = rendering.fill_coords(
            img, rendering.point_in_rect(0, 0.031, 0, 1), COLORS["grey"]
        )
        img = rendering.fill_coords(
            img, rendering.point_in_rect(0, 1, 0, 0.031), COLORS["grey"]
        )

        img = OvercookedV2Visualizer._render_cell(obj, img)

        img_highlight = rendering.highlight_img(img)
        img = jax.lax.select(highlight, img_highlight, img)

        # Downsample the image to perform supersampling/anti-aliasing
        img = rendering.downsample(img, self.subdivs)

        # Cache the rendered tile
        # OvercookedV2Visualizer.tile_cache[key] = img

        return img

    def _render_grid(
        self,
        grid,
        highlight_mask,
    ):
        img_grid = jax.vmap(jax.vmap(self._render_tile))(grid, highlight_mask)

        # print("img_grid", img_grid.shape)

        grid_rows, grid_cols, tile_height, tile_width, channels = img_grid.shape

        big_image = img_grid.transpose(0, 2, 1, 3, 4).reshape(
            grid_rows * tile_height, grid_cols * tile_width, channels
        )

        # print("big_image", big_image.shape)

        return big_image

    def close(self):
        self.window.close()



import jax
import jax.numpy as jnp
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
from jaxmarl.environments.spaces import Discrete, MultiDiscrete
import chex
from flax import struct
from typing import Tuple, Dict
from functools import partial


@struct.dataclass
class State:
    bulb_state: bool  # light on or off
    agent_in_room: int  # index of the agent in room
    has_been: chex.Array  # [bool]*n_agents
    done: bool  # done is the same for all the agents
    step: int


class SwitchRiddle(MultiAgentEnv):
    def __init__(
        self,
        num_agents=3,
        initial_bulb_state=False,
        reward_all_die=-1.0,
        reward_all_live=1.0,
    ):
        assert num_agents >= 3, "The minimum number of agents for this environment is 3"
        self.num_agents = num_agents
        self.agent_range = jnp.arange(num_agents)
        self.agents = [f"agent_{i}" for i in range(num_agents)]
        self.a_to_i = {a: i for i, a in enumerate(self.agents)}

        # action space
        self.game_actions = {
            "NOTHING": 0,
            "SWITCH_LIGHT": 1,
            "TELL": 2,
        }
        self.game_actions_idx = {v: k for k, v in self.game_actions.items()}
        self.action_spaces = {i: Discrete(len(self.game_actions)) for i in self.agents}

        # observation space is: IN_ROOM (1/0) and LIGHT STATE (1/0)
        self.observation_spaces = {i: MultiDiscrete([2, 2]) for i in self.agents}

        # Parameters
        self.max_steps = 4 * self.num_agents - 6
        self.initial_bulb_state = initial_bulb_state
        self.reward_all_die = reward_all_die
        self.reward_all_live = reward_all_live

    @partial(jax.jit, static_argnums=[0])
    def reset(self, key: chex.PRNGKey) -> Tuple[chex.Array, State]:
        state = State(
            bulb_state=jnp.array(self.initial_bulb_state),
            agent_in_room=jax.random.randint(
                key, shape=(), minval=0, maxval=self.num_agents
            ),
            has_been=jnp.full(self.num_agents, 0),
            done=jnp.array(False),
            step=0,
        )

        return self.get_obs(state), state

    @partial(jax.jit, static_argnums=[0])
    def step_env(self, key: chex.PRNGKey, state: State, actions: dict):
        # get the actions as array
        actions = jnp.array([actions[i] for i in self.agents])

        # the relevant action is only the one of the current agent in the room
        bulb_state = state.bulb_state
        agent_in_room = state.agent_in_room
        agent_action = actions.at[
            agent_in_room
        ].get()  # the relevant action is the one of the agent in the room
        has_been = state.has_been.at[agent_in_room].set(1)  # current agent is in room
        done = state.done
        step = state.step + 1

        # check if to change the state of the light
        bulb_state = jnp.where(
            agent_action == self.game_actions["SWITCH_LIGHT"],
            ~state.bulb_state,  # change the state of the light
            state.bulb_state,  # do nothing
        ).squeeze()

        # get the reward if agent spoke
        reward = jnp.where(
            agent_action == self.game_actions["TELL"],
            jnp.where(
                jnp.all(has_been),
                self.reward_all_live,  # positive reward if agent spoke and all agents have been in room
                self.reward_all_die,  # negative reward if agent spoke and not all agents have been in room
            ),
            0.0,  # no reward if agent didn't speak
        ).squeeze()

        # done if an agent spoke or maximum step reached
        done = jnp.logical_or(
            agent_action == self.game_actions["TELL"], state.step + 1 >= self.max_steps
        ).squeeze()

        # update the environment internal state
        state = State(
            bulb_state=bulb_state,
            agent_in_room=jax.random.randint(
                key, shape=(), minval=0, maxval=self.num_agents
            ),
            has_been=has_been,
            done=done,
            step=step,
        )

        # prepare outputs
        obs = self.get_obs(state)
        rewards = {a: reward for a in self.agents}
        dones = {a: done for a in self.agents + ["__all__"]}
        info = {}

        return obs, state, rewards, dones, info

    @partial(jax.jit, static_argnums=[0])
    def get_obs(self, state: State) -> dict:
        @partial(jax.vmap, in_axes=[0, None])
        def _observation(aidx: int, state: State) -> jnp.ndarray:
            """Return observation for agent i."""
            #print('state agent in room', state.agent_in_room)
            agent_in_room = (state.agent_in_room == aidx)
            #print('agent in room', agent_in_room)
            bulb_state = jnp.logical_and(
                agent_in_room, # only the current agent in the room can see the bulb state
                state.bulb_state
            ).squeeze()
            #print('agent shape', agent_in_room.shape, 'bulb state', bulb_state.shape)
            return jnp.array([agent_in_room, bulb_state]).astype(int)

        obs = _observation(self.agent_range, state)
        #print('obs', obs.shape)
        return {a: obs[i] for i, a in enumerate(self.agents)}

    def render(self, state: State):
        print(f"\nCurrent step: {state.step}")
        print(f"Bulb state: {'On' if state.bulb_state else 'Off'}")
        print(f"Agent in room: {state.agent_in_room}")
        print(f"Agents been in room: {state.has_been}")
        print(f"Done: {state.done}")

    @property
    def name(self) -> str:
        """Environment name."""
        return type(self).__name__


def example():
    num_agents = 5
    key = jax.random.PRNGKey(0)

    from jaxmarl import make
    env = make('switch_riddle', num_agents=num_agents)

    obs, state = env.reset(key)
    env.render(state)

    for _ in range(20):
        key, key_reset, key_act, key_step = jax.random.split(key, 4)

        env.render(state)
        print("obs:", obs)

        # Sample random actions.
        key_act = jax.random.split(key_act, env.num_agents)
        actions = {
            agent: env.action_space(agent).sample(key_act[i])
            for i, agent in enumerate(env.agents)
        }

        print(
            "action:",
            env.game_actions_idx[actions[env.agents[state.agent_in_room]].item()],
        )

        # Perform the step transition.
        obs, state, reward, done, infos = env.step(key_step, state, actions)

        print("reward:", reward["agent_0"])


if __name__ == "__main__":
    with jax.disable_jit():
        example()



""" Default parameters for MPE environments """

# Action types
DISCRETE_ACT = "Discrete"
CONTINUOUS_ACT = "Continuous"

# Environment
MAX_STEPS = 25
DAMPING = 0.25
CONTACT_FORCE = 1e2
CONTACT_MARGIN = 1e-3
DT = 0.1

# Colours
AGENT_COLOUR = (115, 243, 115)
ADVERSARY_COLOUR = (243, 115, 115)
OBS_COLOUR = (64, 64, 64)



import jax
import jax.numpy as jnp
from PIL import Image
from jaxmarl import make
from pathlib import Path
import math

agent_pop_sizes = [10]
for n_a in agent_pop_sizes:
    # load environment
    num_agents=n_a
    grid_size=math.ceil(math.log(num_agents+2, 2)**2)
    num_coins=2*num_agents
    num_inner_steps=128
    num_outer_steps=1
    rng = jax.random.PRNGKey(123)
    env = make('storm_np',
            num_inner_steps=num_inner_steps,
            num_outer_steps=num_outer_steps,
            num_agents=num_agents,
            fixed_coin_location=True,
            payoff_matrix=jnp.array([[[3, 0], [5, 1]], [[3, 5], [0, 1]]]),
            freeze_penalty=5,
            grid_size=grid_size,
            num_coins=num_coins
        )
    rng, _rng = jax.random.split(rng)

    root_dir = f"tests/a{num_agents}_g{grid_size}_c{num_coins}_i{num_inner_steps}_o{num_outer_steps}"
    path = Path(root_dir + "/state_pics")
    path.mkdir(parents=True, exist_ok=True)

    for o_t in range(num_outer_steps):
        obs, old_state = env.reset(_rng)

        # render each timestep
        pics = []
        pics1 = []
        pics2 = []

        img = env.render(old_state)
        Image.fromarray(img).save(f"{root_dir}/state_pics/init_state.png")
        pics.append(img)

        for t in range(num_inner_steps):

            rng, *rngs = jax.random.split(rng, num_agents+1)
            actions = [jax.random.choice(
                rngs[a],
                a=env.action_space(0).n,
                p=jnp.array([0.15, 0.15, 0.35, 0.1, 0.1, 0.05, 0.05, 0.05])
            ) for a in range(num_agents)]

            obs, state, reward, done, info = env.step_env(
                rng, old_state, [a for a in actions]
            )

            print('###################')
            print(f'timestep: {t} to {t+1}')
            print(f'actions: {[action.item() for action in actions]}')
            print(f'Defection rate: {info["coin_ratio"]*100}%')
            print("###################")

            img = env.render(state)
            Image.fromarray(img).save(
                f"{root_dir}/state_pics/state_{t+1}.png"
            )
            pics.append(img)

            old_state = state

        # create and save gif
        print("Saving GIF")
        pics = [Image.fromarray(img) for img in pics]
        pics[0].save(
        f"{root_dir}/state_outer_step_{o_t+1}.gif",
        format="GIF",
        save_all=True,
        optimize=False,
        append_images=pics[1:],
        duration=200,
        loop=0,
        )


"""
Test the environments to ensure they implement the base API correctly.
STORM is not included as its observation space does not follow the base API.
"""

import pytest
import numpy as np
import jax
import jax.numpy as jnp

# Import the base environment class
from jaxmarl.environments import MultiAgentEnv
from jaxmarl.environments.spaces import Space
# Import the specific environment to test. Replace 'MyEnv' with your environment's class.

from jaxmarl.registration import make

envs_to_test = [
    "MPE_simple_v3",
    "MPE_simple_tag_v3",
    "MPE_simple_world_comm_v3",
    "MPE_simple_spread_v3",
    "MPE_simple_crypto_v3",
    "MPE_simple_speaker_listener_v4",
    "MPE_simple_push_v3",
    "MPE_simple_adversary_v3",
    "MPE_simple_reference_v3",
    "MPE_simple_facmac_v1",
    "MPE_simple_facmac_3a_v1",
    "MPE_simple_facmac_6a_v1",
    "MPE_simple_facmac_9a_v1",
    "switch_riddle",
    "SMAX",
    "HeuristicEnemySMAX",
    "ant_4x2",
    "halfcheetah_6x1",
    "hopper_3x1",
    "humanoid_9|8",
    "walker2d_2x3",
    # "storm",
    # "storm_2p",
    # "storm_np",
    "hanabi",
    "overcooked",
    "overcooked_v2",
    "coin_game",
    "jaxnav",
]

# A pytest fixture to instantiate the environment.
@pytest.fixture(scope="module", params=envs_to_test)
def env(request):
    env_id = request.param
    env_instance = make(env_id)
    yield env_instance

def test_inherits_base_env(env: MultiAgentEnv):
    """Test that the environment is a subclass of MultiAgentEnv."""
    
    assert isinstance(env, MultiAgentEnv), "Environment does not inherit from MultiAgentEnv"

def test_observation_space_definition(env: MultiAgentEnv):
    assert hasattr(env, "observation_spaces"), "`observation_spaces` does  not exist"

    for agent in env.observation_spaces.keys():
        assert env.observation_spaces[agent] == env.observation_space(agent)
        assert isinstance(env.observation_spaces[agent], Space), f"Agent observation space {env.observation_spaces[agent]} is not a valid JaxMARL space"

def test_action_space_definition(env: MultiAgentEnv):
    assert hasattr(env, "action_spaces"), "`action_spaces` does  not exist"

    for agent in env.action_spaces.keys():
        assert env.action_spaces[agent] == env.action_space(agent)
        assert isinstance(env.action_spaces[agent], Space)

def test_reset_returns_valid_observation(env: MultiAgentEnv):
    """Test that reset() returns an observation that is valid according to observation_space."""
    # Check that the environment defines an observation_space attribute.
    rng = jax.random.PRNGKey(0)

    initial_obs, _ = env.reset(rng)
    # Verify that the returned observation is contained in the observation space.
    for agent in env.observation_spaces.keys():
        print(f"Agent: {agent}, Observation: {initial_obs[agent]}, Space: {env.observation_spaces[agent]}")
        assert env.observation_spaces[agent].contains(initial_obs[agent]), f"Initial observation for agent {agent}, {env.observation_spaces[agent]} does not contain the observation value {initial_obs[agent]}"


def test_step_returns_correct_format(env):
    """Test that step(action) returns a tuple (observation, reward, done, info)
       with valid types and that the observation adheres to the observation_space."""
    # Ensure the environment defines an action_space attribute.
    assert hasattr(env, "action_space"), "Environment missing action_space attribute"
    
    rng = jax.random.PRNGKey(0)

    # Reset the environment first.
    rng, _rng = jax.random.split(rng)
    _, initial_state = env.reset(_rng)
    # Sample a valid action.

    rng, _rng = jax.random.split(rng)
    actions = {
        a: env.action_space(a).sample(_rng) for a in env.agents
    }
    
    # Take a step in the environment.
    rng, _rng = jax.random.split(rng)
    result = env.step(_rng, initial_state, actions)
    # Check that the result is a 4-tuple.
    assert isinstance(result, tuple) and len(result) == 5, "Step did not return a 5-tuple"
    
    next_obs, next_state, reward, done, info = result
    for agent in env.observation_spaces.keys():
        assert env.observation_spaces[agent].contains(next_obs[agent])

    # Check that the reward is numeric.
    assert isinstance(reward, dict), "Reward is not a dictionary"
    # Verify that 'done' is a boolean.
    assert isinstance(done, dict), "Done flag is not a dictionary"
    assert "__all__" in done

    # Check that 'info' is a dictionary.
    assert isinstance(info, dict), "Info is not a dictionary"

@pytest.mark.parametrize("env_id", envs_to_test)
def test_envs(env_id):
    env = make(env_id)
    test_inherits_base_env(env)
    test_observation_space_definition(env)
    test_action_space_definition(env)
    test_reset_returns_valid_observation(env)
    test_step_returns_correct_format(env)

if __name__ == "__main__":
    for env_id in envs_to_test:
        env_instance = make(env_id)
        print(f"Testing environment: {env_id}")
        test_inherits_base_env(env_instance)
        test_observation_space_definition(env_instance)
        test_action_space_definition(env_instance)
        test_reset_returns_valid_observation(env_instance)
        test_step_returns_correct_format(env_instance)



import os
import jax
import numpy as np
from jax import numpy as jnp
from jaxmarl import make
from jaxmarl.wrappers.baselines import load_params, LogWrapper
from jaxmarl.environments.hanabi.pretrained.obl_r2d2_agent import OBLAgentR2D2
import json

seed = 0
n_test_games = 1000
time_limit = 80
env = LogWrapper(make('hanabi'))
agent = OBLAgentR2D2()

batchify = lambda x: jnp.stack([x[a] for a in env.agents])
unbatchify = lambda x: {a:x[i] for i,a in enumerate(env.agents)}

@jax.jit
def run_obl_test(rng, params):

    def _env_step(carry, _):
    
        rng, env_state, agent_carry, last_obs = carry 

        agent_input = (
            batchify(last_obs),
            batchify(env.get_legal_moves(env_state.env_state))
        )

        new_agent_carry, actions = agent.greedy_act(params, agent_carry, agent_input)
        actions = unbatchify(actions)
        
        rng, _rng = jax.random.split(rng)
        new_obs, new_env_state, rewards, dones, infos = env.step(_rng, env_state, actions)

        return (rng, new_env_state, new_agent_carry, new_obs), (infos, rewards, dones)
    
    init_obs, env_state = env.reset(rng)
    agent_carry = agent.initialize_carry(jax.random.PRNGKey(0), batch_dims=(2,))

    _, (infos, rewards, dones) = jax.lax.scan(
        _env_step, (rng, env_state, agent_carry, init_obs), None, time_limit
    )

    # compute the metrics of the first episode that is done for each parallel env
    def first_episode_returns(rewards, dones):
        first_done = jax.lax.select(jnp.argmax(dones)==0., dones.size, jnp.argmax(dones))
        first_episode_mask = jnp.where(jnp.arange(dones.size) <= first_done, True, False)
        return jnp.where(first_episode_mask, rewards, 0.).sum()
        
    cum_rewards = first_episode_returns(rewards['__all__'], dones['__all__'])

    first_returned_episode = jnp.nonzero(infos['returned_episode'], size=1)[0][0]
    returns = infos['returned_episode_returns'][first_returned_episode][0]

    returns = jnp.where(
        first_returned_episode.any(), returns, cum_rewards
    )

    return returns

def main():
        
    rng = jax.random.PRNGKey(seed)
    test_rngs = jax.random.split(rng, n_test_games)
    f = jax.jit(jax.vmap(run_obl_test, in_axes=[0,None]))

    model_dir = './obl-r2d2-flax'
    for d in sorted(os.listdir(model_dir)):
        if not os.path.isdir(os.path.join(model_dir, d)):
            continue
        print(d)
        for file in sorted(os.listdir(os.path.join(model_dir, d))):
            if file.split('.')[-1] == 'safetensors':
                params = load_params(os.path.join(model_dir, d, file))
                returns = f(test_rngs, params)
                print('  ', file.split('.')[0], returns.mean())


if __name__ == '__main__':
    main()



import jax
import jax.numpy as jnp
from typing import NamedTuple
from typing import Tuple
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
import chex
from jaxmarl.environments import spaces
from typing import Optional, Tuple

@chex.dataclass
class EnvState:
    red_pos: jnp.ndarray
    blue_pos: jnp.ndarray
    red_coin_pos: jnp.ndarray
    blue_coin_pos: jnp.ndarray
    inner_t: int
    outer_t: int
    # stats
    red_coop: jnp.ndarray
    red_defect: jnp.ndarray
    blue_coop: jnp.ndarray
    blue_defect: jnp.ndarray
    counter: jnp.ndarray  # 9
    coop1: jnp.ndarray  # 9
    coop2: jnp.ndarray  # 9
    last_state: jnp.ndarray  # 2


STATES = jnp.array(
    [
        [0],  # SS
        [1],  # CC
        [2],  # CD
        [3],  # DC
        [4],  # DD
        [5],  # SC
        [6],  # SD
        [7],  # CS
        [8],  # DS
    ]
)
MOVES = jnp.array(
    [
        [0, 1],  # right
        [0, -1],  # left
        [1, 0],  # up
        [-1, 0],  # down
        [0, 0],  # stay
    ]
)


class CoinGame(MultiAgentEnv):
    """
    JAX Compatible version of coin game environment.
    """

    def __init__(
        self,
        num_inner_steps: int = 10,
        num_outer_steps: int = 10,
        cnn: bool = False,
        egocentric: bool = False,
        shared_rewards: bool = False,
        payoff_matrix=[[1, 1, -2], [1, 1, -2]],
    ):

        super().__init__(num_agents=2)
        self.agents = [str(i) for i in list(range(2))]
        self.payoff_matrix = payoff_matrix
        self.cnn = cnn
        
        # helper functions
        def _update_stats(
            state: EnvState,
            rr: jnp.ndarray,
            rb: jnp.ndarray,
            br: jnp.ndarray,
            bb: jnp.ndarray,
        ):
            def state2idx(s: jnp.ndarray) -> int:
                idx = 0
                idx = jnp.where((s == jnp.array([1, 1])).all(), 1, idx)
                idx = jnp.where((s == jnp.array([1, 2])).all(), 2, idx)
                idx = jnp.where((s == jnp.array([2, 1])).all(), 3, idx)
                idx = jnp.where((s == jnp.array([2, 2])).all(), 4, idx)
                idx = jnp.where((s == jnp.array([0, 1])).all(), 5, idx)
                idx = jnp.where((s == jnp.array([0, 2])).all(), 6, idx)
                idx = jnp.where((s == jnp.array([2, 0])).all(), 7, idx)
                idx = jnp.where((s == jnp.array([1, 0])).all(), 8, idx)
                return idx

            # actions are X, C, D
            a1 = 0
            a1 = jnp.where(rr, 1, a1)
            a1 = jnp.where(rb, 2, a1)

            a2 = 0
            a2 = jnp.where(bb, 1, a2)
            a2 = jnp.where(br, 2, a2)

            # if we didn't get a coin this turn, use the last convention
            convention_1 = jnp.where(a1 > 0, a1, state.last_state[0])
            convention_2 = jnp.where(a2 > 0, a2, state.last_state[1])

            idx = state2idx(state.last_state)
            counter = state.counter + jnp.zeros_like(
                state.counter, dtype=jnp.int16
            ).at[idx].set(1)
            coop1 = state.coop1 + jnp.zeros_like(
                state.counter, dtype=jnp.int16
            ).at[idx].set(rr)
            coop2 = state.coop2 + jnp.zeros_like(
                state.counter, dtype=jnp.int16
            ).at[idx].set(bb)
            convention = jnp.stack([convention_1, convention_2]).reshape(2)
            return counter, coop1, coop2, convention

        def _abs_position(state: EnvState) -> jnp.ndarray:
            obs1 = jnp.zeros((3, 3, 4), dtype=jnp.int8)
            obs2 = jnp.zeros((3, 3, 4), dtype=jnp.int8)

            # obs channels are [red_player, blue_player, red_coin, blue_coin]
            obs1 = obs1.at[state.red_pos[0], state.red_pos[1], 0].set(1)
            obs1 = obs1.at[state.blue_pos[0], state.blue_pos[1], 1].set(1)
            obs1 = obs1.at[
                state.red_coin_pos[0], state.red_coin_pos[1], 2
            ].set(1)
            obs1 = obs1.at[
                state.blue_coin_pos[0], state.blue_coin_pos[1], 3
            ].set(1)

            # each agent has egotistic color (so thinks they are red)
            obs2 = jnp.stack(
                [obs1[:, :, 1], obs1[:, :, 0], obs1[:, :, 3], obs1[:, :, 2]],
                axis=-1,
            )
            obs = {self.agents[0]: obs1, self.agents[1]: obs2}
            return obs

        def _relative_position(state: EnvState) -> jnp.ndarray:
            """Assume canonical agent is red player"""
            # (x) redplayer at (2, 2)
            # (y) redcoin at   (0 ,0)
            #
            #  o o x        o o y
            #  o o o   ->   o x o
            #  y o o        o o o
            #
            # redplayer goes to (1, 1)
            # redcoing goes to  (2, 2)
            # offset = (-1, -1)
            # new_redcoin = (0, 0) + (-1, -1) = (-1, -1) mod3
            # new_redcoin = (2, 2)

            agent_loc = jnp.array([state.red_pos[0], state.red_pos[1]])
            ego_offset = jnp.ones(2, dtype=jnp.int8) - agent_loc

            rel_other_player = (state.blue_pos + ego_offset) % 3
            rel_red_coin = (state.red_coin_pos + ego_offset) % 3
            rel_blue_coin = (state.blue_coin_pos + ego_offset) % 3

            # create observation
            obs = jnp.zeros((3, 3, 4), dtype=jnp.int8)
            obs = obs.at[1, 1, 0].set(1)
            obs = obs.at[rel_other_player[0], rel_other_player[1], 1].set(1)
            obs = obs.at[rel_red_coin[0], rel_red_coin[1], 2].set(1)
            obs = obs.at[rel_blue_coin[0], rel_blue_coin[1], 3].set(1)
            return obs

        def _state_to_obs(state: EnvState) -> jnp.ndarray:
            if egocentric:
                obs1 = _relative_position(state)

                # flip red and blue coins for second agent
                obs2 = _relative_position(
                    EnvState(
                        red_pos=state.blue_pos,
                        blue_pos=state.red_pos,
                        red_coin_pos=state.blue_coin_pos,
                        blue_coin_pos=state.red_coin_pos,
                        inner_t=0,
                        outer_t=0,
                        red_coop=state.blue_coop,
                        red_defect=state.blue_defect,
                        blue_coop=state.red_coop,
                        blue_defect=state.red_defect,
                        last_state=state.last_state,
                        counter=state.counter,
                        coop1=state.coop1,
                        coop2=state.coop2,
                    )
                )
                obs = (obs1, obs2)
                obs = {agent: obs for agent, obs in zip(self.agents, obs)}
            else:
                obs = _abs_position(state)

            if not cnn:
                return {agent: obs[agent].flatten() for agent in obs}
            return obs
        
        _shape = (3, 3, 4) if self.cnn else (36,)
        self.observation_spaces = {
            a: spaces.Box(low=0, high=1, shape=_shape, dtype=jnp.uint8) for a in self.agents
        }
        self.action_spaces = {
            a: spaces.Discrete(5) for a in self.agents
        }    

        def _step(
            key: chex.PRNGKey,
            state: EnvState,
            actions: Tuple[int, int],
        ):
            action_0, action_1 = list(actions.values())
            new_red_pos = (state.red_pos + MOVES[action_0]) % 3
            new_blue_pos = (state.blue_pos + MOVES[action_1]) % 3
            red_reward, blue_reward = 0, 0

            red_red_matches = jnp.all(
                new_red_pos == state.red_coin_pos, axis=-1
            )
            red_blue_matches = jnp.all(
                new_red_pos == state.blue_coin_pos, axis=-1
            )

            blue_red_matches = jnp.all(
                new_blue_pos == state.red_coin_pos, axis=-1
            )
            blue_blue_matches = jnp.all(
                new_blue_pos == state.blue_coin_pos, axis=-1
            )

            ### [[1, 1, -2],[1, 1, -2]]
            _rr_reward = self.payoff_matrix[0][0]
            _rb_reward = self.payoff_matrix[0][1]
            _r_penalty = self.payoff_matrix[0][2]
            _br_reward = self.payoff_matrix[1][0]
            _bb_reward = self.payoff_matrix[1][1]
            _b_penalty = self.payoff_matrix[1][2]

            red_reward = jnp.where(
                red_red_matches, red_reward + _rr_reward, red_reward
            )
            red_reward = jnp.where(
                red_blue_matches, red_reward + _rb_reward, red_reward
            )
            red_reward = jnp.where(
                blue_red_matches, red_reward + _r_penalty, red_reward
            )

            blue_reward = jnp.where(
                blue_red_matches, blue_reward + _br_reward, blue_reward
            )
            blue_reward = jnp.where(
                blue_blue_matches, blue_reward + _bb_reward, blue_reward
            )
            blue_reward = jnp.where(
                red_blue_matches, blue_reward + _b_penalty, blue_reward
            )

            (counter, coop1, coop2, last_state) = _update_stats(
                state,
                red_red_matches,
                red_blue_matches,
                blue_red_matches,
                blue_blue_matches,
            )

            key, subkey = jax.random.split(key)
            new_random_coin_poses = jax.random.randint(
                subkey, shape=(2, 2), minval=0, maxval=3
            )
            new_red_coin_pos = jnp.where(
                jnp.logical_or(red_red_matches, blue_red_matches),
                new_random_coin_poses[0],
                state.red_coin_pos,
            )
            new_blue_coin_pos = jnp.where(
                jnp.logical_or(red_blue_matches, blue_blue_matches),
                new_random_coin_poses[1],
                state.blue_coin_pos,
            )

            next_red_coop = state.red_coop + jnp.zeros(
                num_outer_steps, dtype=jnp.int8
            ).at[state.outer_t].set(red_red_matches)
            next_red_defect = state.red_defect + jnp.zeros(
                num_outer_steps, dtype=jnp.int8
            ).at[state.outer_t].set(red_blue_matches)
            next_blue_coop = state.blue_coop + jnp.zeros(
                num_outer_steps, dtype=jnp.int8
            ).at[state.outer_t].set(blue_blue_matches)
            next_blue_defect = state.blue_defect + jnp.zeros(
                num_outer_steps, dtype=jnp.int8
            ).at[state.outer_t].set(blue_red_matches)

            next_state = EnvState(
                red_pos=new_red_pos,
                blue_pos=new_blue_pos,
                red_coin_pos=new_red_coin_pos,
                blue_coin_pos=new_blue_coin_pos,
                inner_t=state.inner_t + 1,
                outer_t=state.outer_t,
                red_coop=next_red_coop,
                red_defect=next_red_defect,
                blue_coop=next_blue_coop,
                blue_defect=next_blue_defect,
                counter=counter,
                coop1=coop1,
                coop2=coop2,
                last_state=last_state,
            )

            obs = _state_to_obs(next_state)

            # now calculate if done for inner or outer episode
            inner_t = next_state.inner_t
            outer_t = next_state.outer_t
            reset_inner = inner_t == num_inner_steps

            # if inner episode is done, return start state for next game
            reset_obs, reset_state = _reset(key)
            next_state = EnvState(
                red_pos=jnp.where(
                    reset_inner, reset_state.red_pos, next_state.red_pos
                ),
                blue_pos=jnp.where(
                    reset_inner, reset_state.blue_pos, next_state.blue_pos
                ),
                red_coin_pos=jnp.where(
                    reset_inner,
                    reset_state.red_coin_pos,
                    next_state.red_coin_pos,
                ),
                blue_coin_pos=jnp.where(
                    reset_inner,
                    reset_state.blue_coin_pos,
                    next_state.blue_coin_pos,
                ),
                inner_t=jnp.where(
                    reset_inner, jnp.zeros_like(inner_t), next_state.inner_t
                ),
                outer_t=jnp.where(reset_inner, outer_t + 1, outer_t),
                red_coop=next_state.red_coop,
                red_defect=next_state.red_defect,
                blue_coop=next_state.blue_coop,
                blue_defect=next_state.blue_defect,
                counter=counter,
                coop1=coop1,
                coop2=coop2,
                last_state=jnp.where(reset_inner, jnp.zeros(2), last_state),
            )

            obs = {agent: obs for agent, obs in zip(self.agents, [jnp.where(reset_inner, reset_obs[i], obs[i]) for i in obs])}

            blue_reward = jnp.where(reset_inner, 0.0, blue_reward)
            red_reward = jnp.where(reset_inner, 0.0, red_reward)

            if shared_rewards:
                # shared reward (social welfare\sum of agents individual rewards)
                rewards = {agent: reward for agent, reward in zip(self.agents, (sum((red_reward, blue_reward)),  sum((red_reward, blue_reward))))}
            else:
                # individual reward
                rewards = {agent: reward for agent, reward in zip(self.agents, (red_reward, blue_reward))}

            dones = {agent: reset_inner for agent in self.agents}
            dones['__all__'] = reset_inner
            
            infos = {}
            return (
                obs,
                next_state,
                rewards,
                dones,
                infos,
            )

        def _reset(
            key: jnp.ndarray
        ) -> Tuple[jnp.ndarray, EnvState]:
            key, subkey = jax.random.split(key)
            all_pos = jax.random.randint(
                subkey, shape=(4, 2), minval=0, maxval=3
            )

            empty_stats = jnp.zeros((num_outer_steps), dtype=jnp.int8)
            state_stats = jnp.zeros(9)

            state = EnvState(
                red_pos=all_pos[0, :],
                blue_pos=all_pos[1, :],
                red_coin_pos=all_pos[2, :],
                blue_coin_pos=all_pos[3, :],
                inner_t=0,
                outer_t=0,
                red_coop=empty_stats,
                red_defect=empty_stats,
                blue_coop=empty_stats,
                blue_defect=empty_stats,
                counter=state_stats,
                coop1=state_stats,
                coop2=state_stats,
                last_state=jnp.zeros(2),
            )
            obs = _state_to_obs(state)
            return obs, state

        # overwrite Gymnax as it makes single-agent assumptions
        self.step = jax.jit(_step)
        self.reset = jax.jit(_reset)
        

        self.step = _step
        self.reset = _reset

    @property
    def name(self) -> str:
        """Environment name."""
        return "CoinGame-v1"

    @property
    def num_actions(self) -> int:
        """Number of actions possible in environment."""
        return 5

    def action_space(self, agent: str) -> spaces.Discrete:
        """Action space of the environment."""
        return self.action_spaces[agent]

    def observation_space(self, agent: str) -> spaces.Box:
        """Observation space of the environment."""
        return self.observation_spaces[agent]

    def state_space(self) -> spaces.Dict:
        """State space of the environment."""
        _shape = (3, 3, 4) if self.cnn else (36,)
        return spaces.Box(low=0, high=1, shape=_shape, dtype=jnp.uint8)

    def render(self, state: EnvState):
        import numpy as np
        from matplotlib.backends.backend_agg import (
            FigureCanvasAgg as FigureCanvas,
        )
        from matplotlib.figure import Figure
        from PIL import Image

        """Small utility for plotting the agent's state."""
        fig = Figure((5, 2))
        canvas = FigureCanvas(fig)
        ax = fig.add_subplot(121)
        ax.imshow(
            np.zeros((3, 3)),
            cmap="Greys",
            vmin=0,
            vmax=1,
            aspect="equal",
            interpolation="none",
            origin="lower",
            extent=[0, 3, 0, 3],
        )
        ax.set_aspect("equal")

        # ax.margins(0)
        ax.set_xticks(jnp.arange(1, 4))
        ax.set_yticks(jnp.arange(1, 4))
        ax.grid()
        red_pos = jnp.squeeze(state.red_pos)
        blue_pos = jnp.squeeze(state.blue_pos)
        red_coin_pos = jnp.squeeze(state.red_coin_pos)
        blue_coin_pos = jnp.squeeze(state.blue_coin_pos)
        ax.annotate(
            "R",
            fontsize=20,
            color="red",
            xy=(red_pos[0], red_pos[1]),
            xycoords="data",
            xytext=(red_pos[0] + 0.5, red_pos[1] + 0.5),
        )
        ax.annotate(
            "B",
            fontsize=20,
            color="blue",
            xy=(blue_pos[0], blue_pos[1]),
            xycoords="data",
            xytext=(blue_pos[0] + 0.5, blue_pos[1] + 0.5),
        )
        ax.annotate(
            "Rc",
            fontsize=20,
            color="red",
            xy=(red_coin_pos[0], red_coin_pos[1]),
            xycoords="data",
            xytext=(red_coin_pos[0] + 0.3, red_coin_pos[1] + 0.3),
        )
        ax.annotate(
            "Bc",
            color="blue",
            fontsize=20,
            xy=(blue_coin_pos[0], blue_coin_pos[1]),
            xycoords="data",
            xytext=(
                blue_coin_pos[0] + 0.3,
                blue_coin_pos[1] + 0.3,
            ),
        )

        ax2 = fig.add_subplot(122)
        ax2.text(0.0, 0.95, "Timestep: %s" % (state.inner_t))
        ax2.text(0.0, 0.75, "Episode: %s" % (state.outer_t))
        ax2.text(
            0.0, 0.45, "Red Coop: %s" % (state.red_coop[state.outer_t].sum())
        )
        ax2.text(
            0.6,
            0.45,
            "Red Defects : %s" % (state.red_defect[state.outer_t].sum()),
        )
        ax2.text(
            0.0, 0.25, "Blue Coop: %s" % (state.blue_coop[state.outer_t].sum())
        )
        ax2.text(
            0.6,
            0.25,
            "Blue Defects : %s" % (state.blue_defect[state.outer_t].sum()),
        )
        ax2.text(
            0.0,
            0.05,
            "Red Total: %s"
            % (
                state.red_defect[state.outer_t].sum()
                + state.red_coop[state.outer_t].sum()
            ),
        )
        ax2.text(
            0.6,
            0.05,
            "Blue Total: %s"
            % (
                state.blue_defect[state.outer_t].sum()
                + state.blue_coop[state.outer_t].sum()
            ),
        )
        ax2.axis("off")
        canvas.draw()
        image = Image.frombytes(
            "RGB",
            fig.canvas.get_width_height(),
            fig.canvas.tostring_rgb(),
        )
        return image


if __name__ == "__main__":
    action = 1
    rng = jax.random.PRNGKey(0)
    env = CoinGame(8, 16, True, False)

    # params = EnvParams(payoff_matrix=[[1, 1, -2], [1, 1, -2]])
    # obs, state = env.reset(rng, params)
    # pics = []

    # for _ in range(16):
    #     rng, rng1, rng2 = jax.random.split(rng, 3)
    #     a1 = jax.random.randint(rng1, (), minval=0, maxval=4)
    #     a2 = jax.random.randint(rng2, (), minval=0, maxval=4)
    #     obs, state, reward, done, info = env.step(
    #         rng, state, (a1 * action, a2 * action), params
    #     )
    #     img = env.render(state)
    #     pics.append(img)

    # pics[0].save(
    #     "test1.gif",
    #     format="gif",
    #     save_all=True,
    #     append_images=pics[1:],
    #     duration=300,
    #     loop=0,
    # )


""" 
Abstract base class for multi agent gym environments with JAX
Based on the Gymnax and PettingZoo APIs
"""

import jax
import jax.numpy as jnp
from typing import Dict
import chex
from functools import partial
from flax import struct
from typing import Tuple, Optional

from jaxmarl.environments.spaces import Space

@struct.dataclass
class State:
    done: chex.Array
    step: int


class MultiAgentEnv(object):
    """Jittable abstract base class for all JaxMARL Environments."""

    def __init__(
        self,
        num_agents: int,
    ) -> None:
        """
        Args:
            num_agents (int): maximum number of agents within the environment, used to set array dimensions
        """
        self.num_agents = num_agents
        self.observation_spaces = dict()
        self.action_spaces = dict()

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[Dict[str, chex.Array], State]:
        """Performs resetting of the environment.

        Args:
            key (chex.PRNGKey): random key

        Returns:
            Observations (Dict[str, chex.Array]): observations for each agent, keyed by agent name
            State (State): environment state
        """
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0,))
    def step(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: Dict[str, chex.Array],
        reset_state: Optional[State] = None,
    ) -> Tuple[Dict[str, chex.Array], State, Dict[str, float], Dict[str, bool], Dict]:
        """Performs step transitions in the environment. Resets the environment if done.
        To control the reset state, pass `reset_state`. Otherwise, the environment will reset using `self.reset`.

        Args:
            key (chex.PRNGKey): random key
            state (State): environment state
            actions (Dict[str, chex.Array]): agent actions, keyed by agent name
            reset_state (Optional[State], optional): Optional environment state to reset to on episode completion. Defaults to None.

        Returns:
            Observations (Dict[str, chex.Array]): next observations
            State (State): next environment state
            Rewards (Dict[str, float]): rewards, keyed by agent name
            Dones (Dict[str, bool]): dones, keyed by agent name:
            Info (Dict): info dictionary
        """

        key, key_reset = jax.random.split(key)
        obs_st, states_st, rewards, dones, infos = self.step_env(key, state, actions)

        if reset_state is None:
            obs_re, states_re = self.reset(key_reset)
        else:
            states_re = reset_state
            obs_re = self.get_obs(states_re)

        # Auto-reset environment based on termination
        states = jax.tree.map(
            lambda x, y: jax.lax.select(dones["__all__"], x, y), states_re, states_st
        )
        obs = jax.tree.map(
            lambda x, y: jax.lax.select(dones["__all__"], x, y), obs_re, obs_st
        )
        return obs, states, rewards, dones, infos

    def step_env(
        self, key: chex.PRNGKey, state: State, actions: Dict[str, chex.Array]
    ) -> Tuple[Dict[str, chex.Array], State, Dict[str, float], Dict[str, bool], Dict]:
        """Environment-specific step transition.
        
        Args:
            key (chex.PRNGKey): random key
            state (State): environment state
            actions (Dict[str, chex.Array]): agent actions, keyed by agent name

        Returns:
            Observations (Dict[str, chex.Array]): next observations
            State (State): next environment state
            Rewards (Dict[str, float]): rewards, keyed by agent name
            Dones (Dict[str, bool]): dones, keyed by agent name:
            Info (Dict): info dictionary
        """
        
        raise NotImplementedError

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        """Applies observation function to state.
        
        Args:
            State (state): Environment state
            
        Returns:
            Observations (Dict[str, chex.Array]): observations keyed by agent names"""
        raise NotImplementedError

    def observation_space(self, agent: str) -> Space:
        """Observation space for a given agent.
        
        Args:
            agent (str): agent name

        Returns:
            space (Space): observation space
        """
        return self.observation_spaces[agent]

    def action_space(self, agent: str) -> Space:
        """Action space for a given agent.
        
        Args:
            agent (str): agent name

        Returns:
            space (Space): action space
        """
        return self.action_spaces[agent]

    @partial(jax.jit, static_argnums=(0,))
    def get_avail_actions(self, state: State) -> Dict[str, chex.Array]:
        """Returns the available actions for each agent.
        
        Args:
            state (State): environment state

        Returns:
            available actions (Dict[str, chex.Array]): available actions keyed by agent name
        """
        raise NotImplementedError

    @property
    def name(self) -> str:
        """Environment name."""
        return type(self).__name__

    @property
    def agent_classes(self) -> dict:
        """Returns a dictionary with agent classes

        Format:
            agent_names: [agent_base_name_1, agent_base_name_2, ...]
        """
        raise NotImplementedError



import dataclasses
from jaxmarl.environments.smax.smax_env import SMAX
from jaxmarl.environments.smax.smax_env import State as SMAXState
from jaxmarl.environments.smax.heuristic_enemy import (
    create_heuristic_policy,
    get_heuristic_policy_initial_state,
)
from jaxmarl.environments.multi_agent_env import MultiAgentEnv
import chex
from typing import Dict, Optional, Tuple
import jax.numpy as jnp
import jax
from flax.struct import dataclass
from functools import partial


@dataclass
class State:
    # underlying jaxmarl env state
    state: ...
    # the enemy policy state. Needed for recurrent policies or
    # remembering details about previous observations for heuristics.
    enemy_policy_state: ...


class EnemySMAX(MultiAgentEnv):
    """Class that presents the SMAX environment as a single-player
    (but still multi-agent) environment. Functions like a wrapper, but
    not linked with any of the wrapper code because that is used differently."""

    def __init__(self, **env_kwargs):
        self._env = SMAX(**env_kwargs)
        # only one team
        self.num_agents = self._env.num_allies
        self.num_enemies = self._env.num_enemies
        # want to provide a consistent API between this and SMAX
        self.num_allies = self._env.num_allies
        self.agents = [f"ally_{i}" for i in range(self.num_agents)]
        self.enemy_agents = [f"enemy_{i}" for i in range(self.num_enemies)]
        self.all_agents = self.agents + self.enemy_agents
        self.observation_spaces = {
            i: self._env.observation_spaces[i] for i in self.agents
        }
        self.action_spaces = {i: self._env.action_spaces[i] for i in self.agents}

    def __getattr__(self, name: str):
        return getattr(self._env, name)

    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key: chex.PRNGKey) -> Tuple[Dict[str, chex.Array], State]:
        key, reset_key = jax.random.split(key)
        obs, state = self._env.reset(reset_key)
        enemy_policy_state = self.get_enemy_policy_initial_state(key)
        new_obs = {agent: obs[agent] for agent in self.agents}
        new_obs["world_state"] = obs["world_state"]
        return new_obs, State(state=state, enemy_policy_state=enemy_policy_state)

    def get_enemy_actions(self, key, enemy_policy_state, enemy_obs):
        raise NotImplementedError

    def get_enemy_policy_initial_state(self, key):
        raise NotImplementedError

    @partial(jax.jit, static_argnums=(0, 4))
    def step_env(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: Dict[str, chex.Array],
        get_state_sequence=False,
    ):
        jaxmarl_state = state.state
        obs = self._env.get_obs(jaxmarl_state)
        enemy_obs = self._env.get_obs_unit_list(jaxmarl_state)
        enemy_obs = jnp.array([enemy_obs[agent] for agent in self.enemy_agents])
        key, action_key = jax.random.split(key)
        enemy_actions, enemy_policy_state = self.get_enemy_actions(
            action_key, state.enemy_policy_state, enemy_obs
        )
        enemy_actions = jnp.array([enemy_actions[i] for i in self.enemy_agents])
        actions = jnp.array([actions[i] for i in self.agents])
        enemy_movement_actions, enemy_attack_actions = (
            self._env._decode_discrete_actions(enemy_actions)
        )
        if self._env.action_type == "continuous":
            cont_actions = jnp.zeros((len(self.all_agents), 4))
            cont_actions = cont_actions.at[: self.num_allies].set(actions)
            key, action_key = jax.random.split(key)
            ally_movement_actions, ally_attack_actions = (
                self._env._decode_continuous_actions(
                    action_key, jaxmarl_state, cont_actions
                )
            )
            ally_movement_actions = ally_movement_actions[: self.num_allies]
            ally_attack_actions = ally_attack_actions[: self.num_allies]
        else:
            ally_movement_actions, ally_attack_actions = (
                self._env._decode_discrete_actions(actions)
            )

        movement_actions = jnp.concatenate(
            [ally_movement_actions, enemy_movement_actions], axis=0
        )
        attack_actions = jnp.concatenate([ally_attack_actions, enemy_attack_actions], axis=0)

        if not get_state_sequence:
            obs, jaxmarl_state, rewards, dones, infos = self._env.step_env_no_decode(
                key,
                jaxmarl_state,
                (movement_actions, attack_actions),
                get_state_sequence=get_state_sequence,
            )
            new_obs = {agent: obs[agent] for agent in self.agents}
            new_obs["world_state"] = obs["world_state"]
            rewards = {agent: rewards[agent] for agent in self.agents}
            all_done = dones["__all__"]
            dones = {agent: dones[agent] for agent in self.agents}
            dones["__all__"] = all_done

            state = state.replace(
                enemy_policy_state=enemy_policy_state, state=jaxmarl_state
            )
            return new_obs, state, rewards, dones, infos
        else:
            states = self._env.step_env_no_decode(
                key,
                jaxmarl_state,
                (movement_actions, attack_actions),
                get_state_sequence=get_state_sequence,
            )
            return states

    @partial(jax.jit, static_argnums=(0,))
    def get_avail_actions(self, state: State):
        avail_actions = self._env.get_avail_actions(state.state)
        return {agent: avail_actions[agent] for agent in self.agents}

    def get_all_unit_obs(self, state: State):
        return self._env.get_obs(state.state)

    def get_obs(self, state: State) -> Dict[str, chex.Array]:
        obs = self.get_all_unit_obs(state)
        return {agent: obs[agent] for agent in self.agents}

    def get_world_state(self, state: State):
        return self._env.get_world_state(state.state)

    def is_terminal(self, state: State):
        return self._env.is_terminal(state.state)

    def expand_state_seq(self, state_seq):
        # TODO jit/scan this
        expanded_state_seq = []

        # TODO this actually can't take a key because recording this key is really hard
        # it's not exposed to the user so we can't ask them to store it. Not a problem
        # for now but will have to get creative in the future potentially.
        for key, state, actions in state_seq:
            # There is a split in the step function of MultiAgentEnv
            # We call split here so that the action key is the same.
            key, _ = jax.random.split(key)
            states = self.step_env(key, state, actions, get_state_sequence=True)
            states = list(map(SMAXState, *dataclasses.astuple(states)))
            viz_actions = {
                agent: states[-1].prev_attack_actions[i]
                for i, agent in enumerate(self.all_agents)
            }

            expanded_state_seq.append((key, state.state, viz_actions))
            expanded_state_seq.extend(
                zip([key] * len(states), states, [viz_actions] * len(states))
            )
            state = state.replace(
                state=state.state.replace(terminal=self.is_terminal(state))
            )
        return expanded_state_seq


class HeuristicEnemySMAX(EnemySMAX):
    def __init__(self, enemy_shoots=True, attack_mode="closest", **env_kwargs):
        super().__init__(**env_kwargs)
        self.enemy_shoots = enemy_shoots
        self.attack_mode = attack_mode
        self.heuristic_policy = create_heuristic_policy(
            self._env, 1, shoot=self.enemy_shoots, attack_mode=self.attack_mode
        )

    def get_enemy_policy_initial_state(self, key):
        return jax.tree.map(
            lambda *xs: jnp.stack(xs),
            *([get_heuristic_policy_initial_state()] * self.num_enemies),
        )

    def get_enemy_actions(self, key, policy_state, enemy_obs):
        heuristic_action_key = jax.random.split(key, num=self.num_enemies)
        enemy_actions, policy_state = jax.vmap(self.heuristic_policy)(
            heuristic_action_key, policy_state, enemy_obs
        )
        enemy_actions = {
            agent: enemy_actions[self._env.agent_ids[agent] - self.num_agents]
            for agent in self.enemy_agents
        }
        return enemy_actions, policy_state


class LearnedPolicyEnemySMAX(EnemySMAX):
    def __init__(self, policy, params, **env_kwargs):
        super().__init__(**env_kwargs)
        self.policy = policy
        self.params = params

    def get_enemy_policy_initial_state(self, key):
        return self.params

    def get_enemy_actions(self, key, policy_state, enemy_obs):
        pi, _ = self.policy.apply(policy_state, enemy_obs)
        enemy_actions = pi.sample(seed=key)
        enemy_actions = {
            agent: enemy_actions[self._env.agent_ids[agent] - self.num_agents]
            for agent in self.enemy_agents
        }
        enemy_actions = {k: v.squeeze() for k, v in enemy_actions.items()}
        return enemy_actions, policy_state



import jax
from functools import partial
import gymnax

class GymnaxToJaxMARL(object):
    """ Use a Gymnax Environment within JaxMARL. Useful for debugging """
    
    num_agents = 1
    agent = "agent"
    agents = [agent]
    
    def __init__(self, env_name: str, env_kwargs: dict = {}):
        self.env_name = env_name 
        self._env, self.env_params = gymnax.make(env_name, **env_kwargs)
        
    @property
    def default_params(self):
        return self.env_params
        
    @partial(jax.jit, static_argnums=(0,))
    def step(self, key, state, actions, params=None):
        print('act', actions[self.agent])
        obs, state, reward, done, info = self._env.step(key, state, actions[self.agent].squeeze(), params)
        obs = {self.agent: obs}
        reward = {self.agent: reward}
        done = {self.agent: done, "__all__": done}
        return obs, state, reward, done, info
    
    @partial(jax.jit, static_argnums=(0,))
    def reset(self, key, params=None):
        obs, state = self._env.reset(key, params)
        obs = {self.agent: obs}
        return obs, state
        
    def observation_space(self, agent: str):
        return self._env.observation_space(self.env_params)
    
    def action_space(self, agent: str):
        return self._env.action_space(self.env_params)
    
    


import numpy as np
import jax.numpy as jnp
import chex
import jax
from enum import IntEnum
from jax.typing import ArrayLike

MAX_INGREDIENTS = 3


class StaticObject(IntEnum):
    EMPTY = 0
    WALL = 1

    # Agents are only included in the observation grid
    AGENT = 2
    SELF_AGENT = 3

    GOAL = 4
    POT = 5
    RECIPE_INDICATOR = 6
    BUTTON_RECIPE_INDICATOR = 7

    PLATE_PILE = 9
    INGREDIENT_PILE_BASE = 10

    @staticmethod
    def is_ingredient_pile(obj):
        return obj >= StaticObject.INGREDIENT_PILE_BASE

    @staticmethod
    def get_ingredient(obj):
        idx = obj - StaticObject.INGREDIENT_PILE_BASE
        return DynamicObject.ingredient(idx)

    @staticmethod
    def ingredient_pile(idx):
        return StaticObject.INGREDIENT_PILE_BASE + idx


class DynamicObject(IntEnum):
    EMPTY = 0
    PLATE = 1 << 0
    COOKED = 1 << 1

    # every ingredient has two unique bit
    BASE_INGREDIENT = 1 << 2

    @staticmethod
    def ingredient(idx):
        return DynamicObject.BASE_INGREDIENT << 2 * idx

    @staticmethod
    def is_ingredient(obj):
        return ((obj >> 2) != 0) & ((obj & DynamicObject.PLATE) == 0)

    @staticmethod
    def ingredient_count(obj):
        initial_val = (obj >> 2, jnp.array(0))

        def _count_ingredients(x):
            obj, count = x
            return (obj >> 2, count + (obj & 0x3))

        _, count = jax.lax.while_loop(
            lambda x: x[0] > 0, _count_ingredients, initial_val
        )
        return count

    @staticmethod
    def get_ingredient_idx_list(obj):
        res = []
        obj >>= 2
        idx = 0
        while obj > 0:
            res += [idx] * (obj & 0x3).item()
            obj >>= 2
            idx += 1
        return res

    @staticmethod
    def get_ingredient_idx_list_jit(obj):

        def _loop_body(carry):
            obj, pos, idx, res = carry
            count = obj & 0x3

            cond = jnp.arange(MAX_INGREDIENTS)
            cond = (cond >= pos) & (res == -1) & (cond < pos + count)

            res = jnp.where(
                cond,
                idx,
                res,
            )

            return (obj >> 2, pos + count, idx + 1, res)

        def _loop_cond(carry):
            obj, pos, _, _ = carry
            return (obj > 0) & (pos < MAX_INGREDIENTS)

        initial_res = jnp.full((MAX_INGREDIENTS,), -1, dtype=jnp.int32)
        carry = (obj >> 2, 0, 0, initial_res)

        val = jax.lax.while_loop(_loop_cond, _loop_body, carry)
        return val[-1]

    @staticmethod
    def get_ingredient_idx(obj):

        def _body_fun(val):
            obj, idx, res = val
            new_res = jax.lax.select(obj & 0x3 != 0, idx, res)
            return (obj >> 2, idx + 1, new_res)

        def _cond_fun(val):
            obj, _, res = val
            return (obj > 0) & (res == -1)

        initial_val = (obj >> 2, 0, -1)
        val = jax.lax.while_loop(_cond_fun, _body_fun, initial_val)
        return val[-1]

    @staticmethod
    def get_recipe_encoding(recipe: ArrayLike):
        ingredients = jax.vmap(DynamicObject.ingredient)(recipe)
        return jnp.sum(ingredients)


class Direction(IntEnum):
    UP = 0
    DOWN = 1
    RIGHT = 2
    LEFT = 3

    @staticmethod
    def opposite(dir):
        opposite_map = jnp.array(
            [Direction.DOWN, Direction.UP, Direction.LEFT, Direction.RIGHT]
        )
        return opposite_map[dir]


ALL_DIRECTIONS = jnp.array(
    [Direction.UP, Direction.DOWN, Direction.RIGHT, Direction.LEFT]
)

DIR_TO_VEC = jnp.array(
    [
        (0, -1),
        (0, 1),
        (1, 0),
        (-1, 0),
    ],
    dtype=jnp.int8,
)


@chex.dataclass
class Position:
    x: jnp.ndarray
    y: jnp.ndarray

    @staticmethod
    def from_tuple(t):
        x, y = t
        return Position(jnp.array([x]), jnp.array([y]))

    def move(self, direction):
        vec = DIR_TO_VEC[direction]
        return Position(x=self.x + vec[0], y=self.y + vec[1])

    def move_in_bounds(self, direction, width, height):
        new_pos = self.move(direction)
        clipped_x = jnp.clip(new_pos.x, 0, width - 1)
        clipped_y = jnp.clip(new_pos.y, 0, height - 1)
        return Position(x=clipped_x, y=clipped_y)

    def checked_move(self, direction, width, height):
        new_pos = self.move(direction)
        clipped_x = jnp.clip(new_pos.x, 0, width - 1)
        clipped_y = jnp.clip(new_pos.y, 0, height - 1)
        return Position(x=clipped_x, y=clipped_y), (clipped_x == new_pos.x) & (
            clipped_y == new_pos.y
        )

    def to_array(self):
        return jnp.stack([self.x, self.y], axis=-1)

    def delta(self, other):
        return jnp.array([self.x - other.x, self.y - other.y])


@chex.dataclass
class Agent:
    pos: Position
    dir: jnp.ndarray
    inventory: jnp.ndarray

    def get_fwd_pos(self):
        return self.pos.move(self.dir)

    @staticmethod
    def from_position(pos):
        return Agent(pos, jnp.array([Direction.UP]), jnp.zeros((1,)))


class Actions(IntEnum):
    # Turn left, turn right, move forward
    right = 0
    down = 1
    left = 2
    up = 3
    stay = 4
    interact = 5


ACTION_TO_DIRECTION = (
    jnp.full((len(Actions),), -1)
    .at[Actions.right]
    .set(Direction.RIGHT)
    .at[Actions.down]
    .set(Direction.DOWN)
    .at[Actions.left]
    .set(Direction.LEFT)
    .at[Actions.up]
    .set(Direction.UP)
)



from .multi_agent_env import MultiAgentEnv, State
from .mpe import (
    SimpleMPE,
    SimpleTagMPE,
    SimpleWorldCommMPE,
    SimpleSpreadMPE,
    SimpleCryptoMPE,
    SimpleSpeakerListenerMPE,
    SimplePushMPE,
    SimpleAdversaryMPE,
    SimpleReferenceMPE,
    SimpleFacmacMPE,
    SimpleFacmacMPE3a,
    SimpleFacmacMPE6a,
    SimpleFacmacMPE9a,
)
from .smax import SMAX, HeuristicEnemySMAX, LearnedPolicyEnemySMAX
from .switch_riddle import SwitchRiddle
from .overcooked import Overcooked, overcooked_layouts
from .overcooked_v2 import OvercookedV2, overcooked_v2_layouts
from .mabrax import Ant, Humanoid, Hopper, Walker2d, HalfCheetah
from .hanabi import Hanabi
from .storm import InTheGrid, InTheGrid_2p, InTheMatrix
from .coin_game import CoinGame
from .jaxnav import JaxNav



from jaxmarl.environments.overcooked_v2.common import StaticObject
import numpy as np
from typing import List, Tuple, Optional
from dataclasses import dataclass
import itertools

# Layouts from Overcooked-AI
cramped_room = """
WWPWW
OA AO
W   W
WBWXW
"""
asymm_advantages = """
WWWWWWWWW
O WXWOW X
W   P   W
W A PA  W
WWWBWBWWW
"""
coord_ring = """
WWWPW
W A P
BAW W
O   W
WOXWW
"""
forced_coord = """
WWWPW
O WAP
OAW W
B W W
WWWXW
"""
counter_circuit = """
WWWPPWWW
W A    W
B WWWW X
W     AW
WWWOOWWW
"""


# Adapted layouts
cramped_room_v2 = """
WWPWW
0A A1
W   R
WBWXW
"""
asymm_advantages_recipes_center = """
WWWWWWWWW
0 WXR01 X
1   P   W
W A PA  W
WWWBWBWWW
"""
asymm_advantages_recipes_right = """
WWWWWWWWW
0 WXW01 X
1   P   R
W A PA  W
WWWBWBWWW
"""
asymm_advantages_recipes_left = """
WWWWWWWWW
0 WXW01 X
1   P   R
R A PA  W
WWWBWBWWW
"""
two_rooms = """
WWWWWB10W
W   W   R
P A W A W
W   W   X
WWWWWWWWW
"""


# Other Layouts
two_rooms_both = """
W01BWB10W
W   W   R
P A W A W
W   W   X
WWWWWWWWW
"""
long_room = """
WWWWWWWWWWWWWWW
B            AP
0             X
WWWWWWWWWWWWWWW
"""
fun_coordination = """
WWWWWWWWW
0   X   2
RA  P  AW
1   B   3
WWWWWWWWW
"""
more_fun_coordination = """
WWWWWWWWW
W   X   W
RA  P  A1
0   P   2
W   B   W
WWWWWWWWW
"""
fun_symmetries_plates = """
WWWWWWW
B  W  0
R APA X
B  W  1
WWWWWWW
"""
fun_symmetries = """
WWWWBWW
2  W  0
R APA X
2  W  1
WWWWBWW
"""
fun_symmetries1 = """
WWWWWBWW
2  WW  0
R AWPA X
2  WW  1
WWWWWBWW
"""
overcookedv2_demo = """
WWPWW
0A A1
L   R
WBWXW
"""


# Extended Cat-Dog Problem Layouts
grounded_coord_simple = """
WW2WWWWW
W  WB  0
R ALPA X
W  WB  1
WW2WWWWW
"""
grounded_coord_ring = """
WWW2R2WWW
W       W
W WWLWW W
2 0   B 2
RAXAP X R
2 1   B 2
W WWLWW W
W       W
WWW2R2WWW
"""


# Test-Time Protocol Formation Layouts
test_time_simple = """
WW2WWWWW
W  WB  0
R AWPA X
W  WB  1
WW2WWWWW
"""
test_time_wide = """
WWXBWW
0 A  0
1    1
WPWPWW
3 A  3
W    W
WWRWWW
"""


# Demo Cook Layouts
demo_cook_simple = """
WWWWWR2W0WW
0      W  B
W     APA X
1      W  B
WWWWWR2W1WW
"""
demo_cook_wide = """
WWWWBXBWWWW
WWW0 A 1WWW
WWWWWPWWWWW
W    A    W
0  W3R3W  0
W1WWWWWWW1W
"""


@dataclass
class Layout:
    # agent positions list of positions, tuples (x, y)
    agent_positions: List[Tuple[int, int]]

    # width x height grid with static items
    static_objects: np.ndarray

    num_ingredients: int

    # If recipe is none, recipes will be sampled from the possible_recipes
    # If possible_recipes is none, all possible recipes with the available ingredients will be considered
    possible_recipes: Optional[List[List[int]]]

    def __post_init__(self):
        if len(self.agent_positions) == 0:
            raise ValueError("At least one agent position must be provided")
        if self.num_ingredients < 1:
            raise ValueError("At least one ingredient must be available")
        if self.possible_recipes is None:
            self.possible_recipes = self._get_all_possible_recipes(self.num_ingredients)

    @property
    def height(self):
        return self.static_objects.shape[0]

    @property
    def width(self):
        return self.static_objects.shape[1]

    @staticmethod
    def _get_all_possible_recipes(num_ingredients: int) -> List[List[int]]:
        """
        Get all possible recipes given the number of ingredients.
        """
        available_ingredients = list(range(num_ingredients)) * 3
        raw_combinations = itertools.combinations(available_ingredients, 3)
        unique_recipes = set(
            tuple(sorted(combination)) for combination in raw_combinations
        )

        return [list(recipe) for recipe in unique_recipes]

    @staticmethod
    def from_string(grid, possible_recipes=None, swap_agents=False):
        """Assumes `grid` is string representation of the layout, with 1 line per row, and the following symbols:
        W: wall
        A: agent
        X: goal
        B: plate (bowl) pile
        P: pot location
        R: recipe of the day indicator
        L: button recipe indicator
        0-9: Ingredient x pile
        ' ' (space) : empty cell

        Depricated:
        O: onion pile - will be interpreted as ingredient 0


        If `recipe` is provided, it should be a list of ingredient indices, max 3 ingredients per recipe
        If `recipe` is not provided, the recipe will be randomized on reset.
        If the layout does not have a recipe indicator, a fixed `recipe` must be provided.

        If `possible_recipes` is provided, it should be a list of lists of ingredient indices, 3 ingredients per recipe.

        Swap agents will swap the positions of the agents in the layout. This is only used for compatibility with the old Overcooked-AI layouts.
        """

        rows = grid.split("\n")

        if len(rows[0]) == 0:
            rows = rows[1:]
        if len(rows[-1]) == 0:
            rows = rows[:-1]

        row_lens = [len(row) for row in rows]
        static_objects = np.zeros((len(rows), max(row_lens)), dtype=int)

        char_to_static_item = {
            " ": StaticObject.EMPTY,
            "W": StaticObject.WALL,
            "X": StaticObject.GOAL,
            "B": StaticObject.PLATE_PILE,
            "P": StaticObject.POT,
            "R": StaticObject.RECIPE_INDICATOR,
            "L": StaticObject.BUTTON_RECIPE_INDICATOR,
        }

        for r in range(10):
            char_to_static_item[f"{r}"] = StaticObject.INGREDIENT_PILE_BASE + r

        agent_positions = []

        num_ingredients = 0
        includes_recipe_indicator = False
        for r, row in enumerate(rows):
            c = 0
            while c < len(row):
                char = row[c]

                if char == "O":
                    char = "0"

                if char == "A":
                    agent_pos = (c, r)
                    agent_positions.append(agent_pos)

                obj = char_to_static_item.get(char, StaticObject.EMPTY)
                static_objects[r, c] = obj

                if StaticObject.is_ingredient_pile(obj):
                    ingredient_idx = obj - StaticObject.INGREDIENT_PILE_BASE
                    num_ingredients = max(num_ingredients, ingredient_idx + 1)

                if obj == StaticObject.RECIPE_INDICATOR:
                    includes_recipe_indicator = True

                c += 1

        # TODO: add some sanity checks - e.g. agent must exist, surrounded by walls, etc.

        if possible_recipes is not None:
            if not isinstance(possible_recipes, list):
                raise ValueError("possible_recipes must be a list")
            if not all(isinstance(recipe, list) for recipe in possible_recipes):
                raise ValueError("possible_recipes must be a list of lists")
            if not all(len(recipe) == 3 for recipe in possible_recipes):
                raise ValueError("All recipes must be of length 3")
        elif not includes_recipe_indicator:
            raise ValueError(
                "Layout does not include a recipe indicator, a fixed recipe must be provided"
            )

        if swap_agents:
            agent_positions = agent_positions[::-1]

        layout = Layout(
            agent_positions=agent_positions,
            static_objects=static_objects,
            num_ingredients=num_ingredients,
            possible_recipes=possible_recipes,
        )

        return layout


overcooked_v2_layouts = {
    # Overcooked-AI layouts
    "cramped_room": Layout.from_string(
        cramped_room, possible_recipes=[[0, 0, 0]], swap_agents=True
    ),
    "asymm_advantages": Layout.from_string(
        asymm_advantages, possible_recipes=[[0, 0, 0]]
    ),
    "coord_ring": Layout.from_string(coord_ring, possible_recipes=[[0, 0, 0]]),
    "forced_coord": Layout.from_string(forced_coord, possible_recipes=[[0, 0, 0]]),
    "counter_circuit": Layout.from_string(
        counter_circuit, possible_recipes=[[0, 0, 0]], swap_agents=True
    ),
    # Adapted layouts
    "cramped_room_v2": Layout.from_string(cramped_room_v2),
    "asymm_advantages_recipes_center": Layout.from_string(
        asymm_advantages_recipes_center
    ),
    "asymm_advantages_recipes_right": Layout.from_string(
        asymm_advantages_recipes_right
    ),
    "asymm_advantages_recipes_left": Layout.from_string(asymm_advantages_recipes_left),
    "two_rooms": Layout.from_string(two_rooms),
    # Other layouts
    "two_rooms_both": Layout.from_string(two_rooms_both),
    "long_room": Layout.from_string(long_room, possible_recipes=[[0, 0, 0]]),
    "fun_coordination": Layout.from_string(
        fun_coordination, possible_recipes=[[0, 0, 2], [1, 1, 3]]
    ),
    "more_fun_coordination": Layout.from_string(
        more_fun_coordination, possible_recipes=[[0, 1, 1], [0, 2, 2]]
    ),
    "fun_symmetries": Layout.from_string(
        fun_symmetries, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    "fun_symmetries_plates": Layout.from_string(
        fun_symmetries_plates, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    "fun_symmetries1": Layout.from_string(
        fun_symmetries1, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    # Extended Cat-Dog Problem Layouts
    "grounded_coord_simple": Layout.from_string(
        grounded_coord_simple, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    "grounded_coord_ring": Layout.from_string(
        grounded_coord_ring, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    # Test-Time Protocol Formation Layouts
    "test_time_simple": Layout.from_string(
        test_time_simple, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    "test_time_wide": Layout.from_string(
        test_time_wide, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    # Demo Cook Layouts
    "demo_cook_simple": Layout.from_string(
        demo_cook_simple, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
    "demo_cook_wide": Layout.from_string(
        demo_cook_wide, possible_recipes=[[0, 0, 0], [1, 1, 1]]
    ),
}



"""
2D robot navigation simulator that follows the JaxMARL interface 
"""

import jax
import jax.numpy as jnp
from functools import partial
import chex 
from flax import struct
from typing import Tuple, Dict
import matplotlib.axes._axes as axes

from jaxmarl.environments import MultiAgentEnv
from jaxmarl.environments.spaces import Box, Discrete

from .maps import make_map, Map
from .jaxnav_utils import wrap, cart2pol

NUM_REWARD_COMPONENTS = 2
REWARD_COMPONENT_SPARSE = 0
REWARD_COMPONENT_DENSE = 1
@struct.dataclass
class Reward:
    sparse: jnp.ndarray
    dense: jnp.ndarray

def listify_reward(reward: Reward, do_batchify=False): # returns shape of (*batch, num_agents, 2)
    ans = jnp.stack(
        [reward.sparse, reward.dense],
        axis=-1
    )
    # batchify stacks the agents first and then does reshape, which is why we need the swapaxes.
    if do_batchify:
        ans = jnp.swapaxes(ans, 0, 1).reshape(-1, *ans.shape[2:]) # shape of (batch * num_agents, 2)
    return ans

@struct.dataclass
class State:
    pos: chex.Array  # [n, [x, y, theta]]
    theta: chex.Array # [n, theta]
    vel: chex.Array  # [n, [speed, omega]]
    done: chex.Array  # [n, bool] whether an agent has terminated
    term: chex.Array  # [n, bool] whether an agent acted in this step
    goal_reached: chex.Array  # [n, bool] whether an agent has reached goal
    move_term: chex.Array  # [n, bool] whether an agent has crashed
    step: int  # step count  
    ep_done: bool  # whether epsiode has terminated
    goal: chex.Array  # [n, x, y]
    map_data: chex.Array  # occupancy grid for environment map
    rew_lambda: float  # linear interpolation between individual and team rewards
    
@struct.dataclass
class EnvInstance:
    agent_pos: chex.Array
    agent_theta: chex.Array
    goal_pos: chex.Array
    map_data: chex.Array
    rew_lambda: chex.Array

### ---- Discrete action constants ----
DISCRETE_ACTS = jnp.array([
    jnp.array([0.0, 0.5]),
    jnp.array([0.0, 0.25]),
    jnp.array([0.0, 0.0]),
    jnp.array([0.0, -0.25]),
    jnp.array([0.0, -0.5]),
    jnp.array([0.5, 0.5]),
    jnp.array([0.5, 0.25]),
    jnp.array([0.5, 0.0]),
    jnp.array([0.5, -0.25]),
    jnp.array([0.5, -0.5]),
    jnp.array([1.0, 0.5]),
    jnp.array([1.0, 0.25]),
    jnp.array([1.0, 0.0]),
    jnp.array([1.0, -0.25]),
    jnp.array([1.0, -0.5]),
], dtype=jnp.float32)


@partial(jax.vmap, in_axes=[0])
def discrete_act_map(action: int) -> jnp.ndarray:
    print('action', action, action.shape)
    return DISCRETE_ACTS[action]
    
## ---- Environment defaults ----
AGENT_BASE = "agent"
MAP_PARAMS = {
    "map_size": (11, 11),
    "fill": 0.3,
}

## ---- Environment ----
class JaxNav(MultiAgentEnv):
    """ 
    Current assumptions:
     - homogenous agents
    """
        
    def __init__(self,
                 num_agents: int = 2, # Number of agents
                 act_type="Continuous", # Action type, either Continuous or Discrete
                 normalise_obs=True,
                 rad=0.3,  # Agent radius, TODO remove dependency on this
                 evaporating=False,  # Whether agents evaporate (dissapeare) when they reach the goal
                 map_id="Grid-Rand-Poly",  # Map type
                 map_params=MAP_PARAMS,  # Map parameters
                 lidar_num_beams=200,
                 lidar_range_resolution=0.05,
                 lidar_max_range=6.0,
                 lidar_min_range=0.0,
                 lidar_angle_factor=1.0,
                 min_v=0.0,
                 max_v=1.0, 
                 max_v_acc=1.0,
                 max_w=0.6,
                 max_w_acc=1.0,
                 max_steps=500,
                 dt=0.1,
                 fixed_lambda=True,
                 rew_lambda=0.5, # linear interpolation between individual and team rewards
                 lambda_range=[0.0, 1.0],
                 goal_radius=0.5,
                 goal_rew=4.0,
                 weight_g=0.25,
                 lim_w=0.7,
                 weight_w=-0.0,
                 dt_rew=-0.01,
                 coll_rew=-4.0,
                 lidar_thresh=0.1,
                 lidar_rew=-0.1,
                 do_sep_reward=False,
                 share_only_sparse=False,
                 info_by_agent=True,
        ):
        super().__init__(num_agents)
        
        assert rad < 1, "current code assumes radius of less than 1"
        self.rad = rad
        self.agents = ["agent_{}".format(i) for i in range(num_agents)]
        self.agent_range = jnp.arange(0, num_agents)
        self.evaporating = evaporating
        
        self._map_obj = make_map(map_id, self.num_agents, self.rad, **map_params)
        self._act_type = act_type
        if self._act_type == "Discrete":
            assert min_v == 0.0, "min_v must be 0.0 for Discrete actions"
        
        # Lidar parameters
        self.normalise_obs = normalise_obs
        self.lidar_num_beams = lidar_num_beams
        self.lidar_max_range = lidar_max_range
        self.lidar_min_range = lidar_min_range
        assert self.lidar_min_range == 0.0, "lidar_min_range must be 0.0" # TODO for now
        self.lidar_range_resolution = lidar_range_resolution
        self.lidar_angle_factor = lidar_angle_factor
        self.lidar_max_angle = jnp.pi * self.lidar_angle_factor
        self.lidar_angles = jnp.linspace(-jnp.pi * self.lidar_angle_factor, jnp.pi * self.lidar_angle_factor, self.lidar_num_beams) 
        num_lidar_samples = int((self.lidar_max_range - self.lidar_min_range) / self.lidar_range_resolution)
        self.lidar_ranges = jnp.linspace(self.lidar_min_range, self.lidar_max_range, num_lidar_samples)
        
        assert min_v < max_v, "min_v must be less than max_v"
        if min_v != 0.0: print(f"WARNING: min_v is not 0.0, it is {min_v}")
        self.min_v = min_v  # min linear velocity (m/s)
        self.max_v = max_v  # max linear velocity (m/s)
        self.max_v_acc = max_v_acc  # max linear acceleration (m/s^2)
        self.max_w = max_w  # max angular velocity (rad/s)
        self.max_w_acc = max_w_acc  # max angular acceleration (rad/s^2)
        self.max_steps = max_steps  # max environment steps within an episode
        self.dt = dt  # seconds per step (s)    

        # Rewards
        # if share_only_sparse:
        #     do_sep_reward, f"If share_only_sparse is True, do_sep_reward must be True for it to work, it is current: {do_sep_reward}"
        self.do_sep_reward = do_sep_reward
        self.share_only_sparse = share_only_sparse
        self.fixed_lambda = fixed_lambda
        self.rew_lambda = rew_lambda  # linear interpolation between individual and team rewards
        self.lambda_range = lambda_range
        if self.fixed_lambda: assert self.rew_lambda is not None, "If fixed_lambda is True, rew_lambda must be set"
        self.goal_radius = goal_radius  # goal radius (m)
        self.goal_rew = goal_rew
        self.weight_g = weight_g
        self.lim_w = lim_w
        self.weight_w = weight_w
        self.dt_rew = dt_rew
        self.coll_rew = coll_rew
        self.lidar_thresh = lidar_thresh 
        self.lidar_rew = lidar_rew
        
        self.info_by_agent = info_by_agent
        self.eval_solved_rate = self.get_eval_solved_rate_fn()
        
        self.action_spaces = {a: self.agent_action_space() for a in self.agents}
        self.observation_spaces = {a: self.agent_observation_space() for a in self.agents}
        
    @property
    def map_obj(self) -> Map:
        """ Return map object """
        return self._map_obj
    
    
    @partial(jax.jit, static_argnums=[0])  
    def reset(self, key: chex.PRNGKey) -> Tuple[Dict[str, chex.Array], State]:
        """ Reset environment. Returns initial agent observations, states and the enviornment state """
        
        state = self.sample_test_case(key)
        obs = self._get_obs(state)
        return {a: obs[i] for i, a in enumerate(self.agents)}, state
        
    @partial(jax.jit, static_argnums=[0])  
    def step_env(
                self,
                key: chex.PRNGKey,
                agent_states: State,
                actions: Dict[str, chex.Array]
        ):
        actions = jnp.array([actions[a] for a in self.agents])  # Batchify
        
        # 1) Update agent states
        if self._act_type == "Discrete": actions = discrete_act_map(actions).reshape(actions.shape[0], 2)
        old_pos = agent_states.pos
        update_state_valid = agent_states.done 
        if not self.evaporating:
            update_state_valid = update_state_valid | agent_states.move_term
        new_pos, new_theta, new_vel = self.update_state(agent_states.pos, agent_states.theta, agent_states.vel, actions, update_state_valid)                      
        step = agent_states.step+1 
                        
        # 2) Check collisions, goal and time
        old_goal_reached = agent_states.goal_reached
        old_move_term = agent_states.move_term
        map_collisions = jax.vmap(self._map_obj.check_agent_map_collision, in_axes=(0, 0, None))(new_pos, new_theta, agent_states.map_data)*(1-agent_states.done).astype(bool)
        if self.num_agents > 1:
            agent_collisions = self.map_obj.check_all_agent_agent_collisions(new_pos, new_theta)*(1- agent_states.done).astype(bool)
        else:
            agent_collisions = jnp.zeros((self.num_agents,), dtype=jnp.bool_)
        collisions = map_collisions | agent_collisions
        goal_reached = (self._check_goal_reached(new_pos, agent_states.goal)*(1-agent_states.done)).astype(bool)
        time_up = jnp.full((self.num_agents,), (step >= self.max_steps))

        # 3) Compute rewards and done values
        old_done = agent_states.done
        if self.evaporating:
            dones = collisions | goal_reached | time_up | agent_states.done  # OR operation over agent status
            ep_done = jnp.all(dones) 
        else:
            goal_reached = goal_reached | old_goal_reached
            collisions = collisions | old_move_term
            ep_done = jnp.all(goal_reached | collisions | time_up)
            dones = jnp.full((self.num_agents,), ep_done)        

        # 4) Update JAX state
        agent_states = agent_states.replace(
            pos=new_pos,
            theta=new_theta,
            vel=new_vel,
            move_term=collisions,
            goal_reached=goal_reached,
            done=dones,
            term=old_done,
            step=step,  
            ep_done=ep_done,
        )
                
        dones = {a: agent_states.done[i] for i, a in enumerate(self.agents)}
        dones["__all__"] = ep_done
        
        # 5) Compute observations
        obs_batch = self._get_obs(agent_states)
        
        # 6) Reward
        rew_individual, individual_rew_sep = self.compute_reward(
            obs_batch,
            agent_states.pos,
            old_pos,
            actions,
            agent_states.goal,
            collisions,
            goal_reached,
            old_done,
            old_goal_reached,
            old_move_term
        )
        avg_rew = rew_individual.mean()
        if self.share_only_sparse:
            shared_rew = individual_rew_sep.sparse.mean()
        else:
            shared_rew = avg_rew
            
        if self.do_sep_reward:
            rew_batch = self.rew_lambda * rew_individual + (1 - self.rew_lambda) * shared_rew
        else:
            rew_batch = self.rew_lambda * rew_individual + (1 - self.rew_lambda) * shared_rew

        rew = {a: rew_batch[i] for i, a in enumerate(self.agents)}
        obs = {a: obs_batch[i] for i, a in enumerate(self.agents)}
        
        if self.evaporating:
            num_c = jnp.sum(collisions | goal_reached | time_up)
            time_o = time_up & ~old_done
        else:
            num_c = jax.lax.select(ep_done, self.num_agents, 0)
            time_o = time_up & ~(collisions | goal_reached)
            
        goal_r = goal_reached * (1 - old_goal_reached)
        agent_c = agent_collisions * (1 - old_move_term)
        map_c = map_collisions * (1 - old_move_term)
        rew_info = avg_rew
        if not self.info_by_agent:
            goal_r = jnp.sum(goal_r)
            agent_c = jnp.sum(agent_c)
            map_c = jnp.sum(map_c)
            time_o = jnp.sum(time_o)
            term = {a: old_done[i] for i, a in enumerate(self.agents)}
        else:
            num_c = jnp.full((self.num_agents,), ep_done, dtype=jnp.int32)
            rew_info = rew_batch
            term = old_done
        
        info = {
            # outcomes
            "NumC": num_c,
            "GoalR": goal_r,
            "AgentC": agent_c,
            "MapC": map_c,
            "TimeO": time_o,
            "Return": rew_info,  # reward
            "terminated": term,  # whether action was valid
        }
        if self.do_sep_reward:
            raise NotImplementedError("Separate reward not implemented")
            return obs, agent_states, individual_rew_sep, dones, info  # NOTE no sharing ..?
        else:
            return obs, agent_states, rew, dones, info
    
    def _lidar_sense(self, idx: int, state: State) -> chex.Array:
        """ Return observation for an agent given the current world state """
        
        pos = state.pos[idx]
        theta = state.theta[idx]
        
        point_fn = jax.vmap(self._map_obj.check_point_map_collision, in_axes=(0, None))
        
        angles = self.lidar_angles + theta

        angles_ranges_mesh = jnp.meshgrid(angles, self.lidar_ranges)  # value mesh
        angles_ranges = jnp.dstack(angles_ranges_mesh)  # reformat array [num_points_per_beam, num_beams, 2]
        beam_coords_x = (angles_ranges[:,:,1]*jnp.cos(angles_ranges[:,:,0])).T + pos[0]  
        beam_coords_y = (angles_ranges[:,:,1]*jnp.sin(angles_ranges[:,:,0])).T + pos[1]
        beam_coords = jnp.dstack((beam_coords_x, beam_coords_y))  # [num_beams, num_points_per_beam, 2]
    
        agent_c = self._lidar_agent_check(self.agent_range, state.pos, state.theta, beam_coords, idx)
        rc_range = jnp.where(agent_c==-1, jnp.inf, agent_c)
        rc_m = jnp.min(rc_range, axis=0)
        rc = jnp.where(rc_m==jnp.inf, -1, rc_m).astype(int)
        
        lidar_hits = point_fn(beam_coords.reshape(-1, 2), state.map_data).reshape(beam_coords.shape[0], beam_coords.shape[1], -1)
        
        idxs = jnp.arange(0, beam_coords.shape[0])
        lidar_hits = lidar_hits.at[idxs, rc].set(1)
        fh_idx = jnp.argmax(lidar_hits>0, axis=1)
        return self.lidar_ranges[fh_idx]
    
    @partial(jax.vmap, in_axes=[None, 0, 0, 0, None, None])
    def _lidar_agent_check(self, other_idx, other_pos, other_theta, beam_coords, host_idx):
        """ Compute lidar collisions with other robots, vectorised across other agent indicies 
        Returns:
            chex.Array: index of lidar particle for which the hit occured, direction to goal in global frame,
                        distance to goal
        """
        
        i = jax.vmap(
            self._map_obj.check_agent_beam_intersect,
            in_axes=(0, None, None, None)
        )(beam_coords, other_pos, other_theta, self.lidar_range_resolution)        
        return jax.lax.select(other_idx==host_idx, jnp.full(i.shape, -1), i)
        
            
    def normalise_lidar(self, ranges):
        return ranges/self.lidar_max_range - 0.5
    
    def unnormalise_lidar(self, ranges):
        return (ranges + 0.5) * self.lidar_max_range
            
    def get_avail_actions(self, state: State):
        
        return {a: jnp.array([1.0, 1.0]) for a in self.agents}  
             
    def sample_test_case(self, key: chex.PRNGKey) -> State:
        
        key_tc, key_lambda = jax.random.split(key)
        map_data, test_case = self._map_obj.sample_test_case(key_tc)
                        
        states = State(
            pos=test_case[:, 0, :2],
            theta=test_case[:, 0, 2], 
            vel=jnp.zeros((self.num_agents, 2)),
            done=jnp.full((self.num_agents,), False),
            term=jnp.full((self.num_agents,), False),  # TODO don't think this is  needed
            goal_reached=jnp.full((self.num_agents,), False),
            move_term=jnp.full((self.num_agents,), False),
            step=0,
            ep_done=False,
            goal=test_case[:, 1, :2],
            map_data=map_data,
            rew_lambda=self.sample_lambda(key_lambda),
        )
        
        return states

    @partial(jax.jit, static_argnums=[0])
    def sample_lambda(self, key):
        if self.fixed_lambda:
            rew_lambda = self.rew_lambda
        else:
            rew_lambda = jax.random.uniform(key, (1,), minval=self.lambda_range[0], maxval=self.lambda_range[1])
        return rew_lambda
    
    @partial(jax.vmap, in_axes=(None, 0, 0, None))
    def _check_map_collisions(self, pos: chex.Array, theta: chex.Array, map_data: chex.Array) -> bool:
        return self._map_obj.check_agent_map_collision(pos, theta, map_data)
       
    @partial(jax.vmap, in_axes=(None, 0, 0))
    def _check_goal_reached(self, pos: chex.Array, goal_pos: chex.Array) -> bool:
        return jnp.sqrt(jnp.sum((pos - goal_pos)**2)) <= self.goal_radius 
            
    @partial(jax.jit, static_argnums=[0])
    def get_obs(self, state: State) -> chex.Array:
        obs_batch = self._get_obs(state)
        return {a: obs_batch[i] for i, a in enumerate(self.agents)}
        
    @partial(jax.jit, static_argnums=[0])
    def _get_obs(self, state: State) -> chex.Array:
        """ Return observation for an agent given the current world state
        
        obs: [lidar (num lidar beams), speeds (2), goal (2), lambda (1)]
        """
         
        @partial(jax.vmap, in_axes=[0, None])
        def _observation(idx: int, state: State) -> jnp.ndarray:
            """Return observation for agent i."""
            
            lidar = self._lidar_sense(idx, state).squeeze()
                    
            vel_obs = state.vel[idx] 
            goal_dir = state.goal[idx] - state.pos[idx]
            goal_obs = cart2pol(*goal_dir)
            goal_dist = jnp.clip(goal_obs[0], 0, self.lidar_max_range) 
            goal_orient = wrap(goal_obs[1]-state.theta[idx])
            
            if self.normalise_obs:
                lidar = self.normalise_lidar(lidar)
                vel_obs = vel_obs / jnp.array([self.max_v, self.max_w]) - jnp.array([0.5, 0.0])
                goal_dist = goal_dist/self.lidar_max_range - 0.5
                goal_orient = goal_orient/jnp.pi
                rew_lambda = state.rew_lambda - 0.5
            vel_goal = jnp.concatenate([vel_obs, goal_dist[None], goal_orient[None], jnp.array([rew_lambda]).reshape(1)])
            return jnp.concatenate((lidar, vel_goal))
        
        return _observation(self.agent_range, state)
    
    @partial(jax.jit, static_argnums=[0])
    def get_world_state(self, state: State) -> chex.Array:
        walls = state.map_data.at[1:-1, 1:-1].get().flatten()
        pos = (state.pos / jnp.array([self._map_obj.width, self._map_obj.height]) - 0.5).flatten()
        theta = (state.theta / jnp.pi - 0.5).flatten()
        goal = (state.goal / jnp.array([self._map_obj.width, self._map_obj.height]) - 0.5).flatten()
        vel = (state.vel / jnp.array([self.max_v, self.max_w]) - 0.5).flatten()
        step = jnp.array(state.step / self.max_steps - 0.5)[None]
        concat = (jnp.concatenate([walls, pos, theta, goal, vel, step])[None]).repeat(self.num_agents, axis=0)
        agent_idx = jnp.eye(self.num_agents)
        
        obs = self._get_obs(state)
        
        return jnp.concatenate([agent_idx, concat, obs], axis=1)
    
    @partial(jax.vmap, in_axes=(None, 0, 0, 0, 0, 0))
    def update_state(self, pos: chex.Array, theta: float, speed: chex.Array, action: chex.Array, done: chex.Array) -> chex.Array:
        """ Update agent's state, if `done` the current position and velocity are returned"""
        if self.evaporating:
            out_done = (jnp.array([0.0,0.0]), theta, jnp.array([0.0,0.0]))  # "Evaporating" agents
        else:
            out_done = (pos, theta, jnp.array([0.0, 0.0]))

        # check if action within limits        
        v_acc = jnp.clip((action[0] - speed[0])/self.dt, -self.max_v_acc, self.max_v_acc)
        w_acc = jnp.clip((action[1] - speed[1])/self.dt, -self.max_w_acc, self.max_w_acc)
        
        v = jnp.clip(speed[0] + v_acc*self.dt, self.min_v, self.max_v)
        w = jnp.clip(speed[1] + w_acc*self.dt, -self.max_w, self.max_w)
        
        dx = v * jnp.cos(theta) * self.dt
        dy = v * jnp.sin(theta) * self.dt
        pos = pos + jnp.array([dx, dy])
        theta = wrap(theta + w*self.dt)
        
        out = (pos, theta, jnp.array([v, w], dtype=jnp.float32))
        return jax.tree.map(lambda x, y: jax.lax.select(done, x, y), out_done, out)  
        
    @partial(jax.vmap, in_axes=(None, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0))
    def compute_reward(
                    self,
                    obs,
                    new_pos,
                    old_pos,
                    act,
                    goal,
                    collision,
                    goal_reached,
                    done,
                    old_goal_reached,
                    old_move_term,
        ):
        rga = self.weight_g * (jnp.linalg.norm(old_pos - goal) - jnp.linalg.norm(new_pos - goal))
        rg = jnp.where(goal_reached, self.goal_rew, rga)  * (1 - old_goal_reached)# goal reward
        rc = collision * self.coll_rew * (1 - old_move_term) # collision reward
        rw = jax.lax.select(jnp.abs(act[1]) > self.lim_w, self.weight_w * jnp.abs(act[1]), 0.0)  # angular velocity magnitue penalty
        rt = self.dt_rew * (1 - (old_goal_reached | old_move_term))  # time penalty
        rl = jnp.any(self.unnormalise_lidar(obs[:self.lidar_num_beams]) <= (self.lidar_thresh + self.rad)) * self.lidar_rew # lidar proximity reward 

        ret = Reward((jnp.where(goal_reached, rg, 0.0) + rc + rt)*(1 - done),
                     (jnp.where(goal_reached, 0.0, rg) + rw + rl)*(1 - done))

        # {
        #     # Sparse reward is goal reward if it was reached & collision reward.
        #     'sparse': jnp.where(goal_reached, rg, 0.0) + rc + rt,
        #     # 
        #     'dense': jnp.where(goal_reached, 0.0, rg) + rw + rl,
        # }

        return (rg + rc + rw + rt + rl)*(1 - done), ret
    
    def set_state(
        self,
        state: State
    ) -> Tuple[Dict[str, chex.ArrayTree], State]:
        """
        Implemented for basic envs.
        """
        obs = self._get_obs(state)
        return {a: obs[i] for i, a in enumerate(self.agents)}, state
        
    def set_env_instance(
        self,
        encoding: EnvInstance
    ) -> Tuple[Dict[str, chex.ArrayTree], State]:
        """
        Instance is encoded as a PyTree containing the following fields:
        agent_pos, agent_theta, goal_pos, map_data
        """
        state = State(
            pos=encoding.agent_pos,
            theta=encoding.agent_theta,
            vel=jnp.zeros((self.num_agents, 2)),
            done=jnp.full((self.num_agents,), False),
            term=jnp.full((self.num_agents,), False),
            goal_reached=jnp.full((self.num_agents,), False),
            move_term=jnp.full((self.num_agents,), False),
            step=0,
            ep_done=False,
            goal=encoding.goal_pos,
            map_data=encoding.map_data,
            rew_lambda=encoding.rew_lambda
        )
        obs = self._get_obs(state)
        return {a: obs[i] for i, a in enumerate(self.agents)}, state
        
    @partial(jax.jit, static_argnums=(0))
    def reset_to_level(self, level: Tuple[chex.Array, chex.Array]) -> Tuple[chex.Array, State]:
        print(' ** WARNING ** reset_to_level in JaxNav is deprecated, use set_state instead')
        map_data, test_case = level
        
        state = State(
            pos=test_case[:, 0, :2],
            theta=test_case[:, 0, 2], 
            vel=jnp.zeros((self.num_agents, 2)),
            done=jnp.full((self.num_agents,), False),
            term=jnp.full((self.num_agents,), False),
            step=0,
            ep_done=False,
            goal=test_case[:, 1, :2],
            map_data=map_data,
            rew_lambda=self.rew_lambda,
        )
        obs = self._get_obs(state)
        return {a: obs[i] for i, a in enumerate(self.agents)}, state
        
    @partial(jax.jit, static_argnums=(0,))
    def step_plr(
        self,
        key: chex.PRNGKey,
        state: State,
        actions: chex.Array,
        level: Tuple,
    ):
        """ Resets to PLR level rather than a random one."""
        print(' ** WARNING ** step_plr in JaxNav is deprecated ')
        obs_st, state_st, rewards, dones, infos = self.step_env(
            key, state, actions
        )
        obs_re, state_re = self.reset_to_level(level)  # todo maybe should be set state depending on PLR code
        state = jax.tree.map(
            lambda x, y: jax.lax.select(state_st.ep_done, x, y), state_re, state_st
        )
        obs = jax.tree.map(
            lambda x, y: jax.lax.select(state_st.ep_done, x, y), obs_re, obs_st
        )
        #obs = jax.lax.select(state_st.ep_done, obs_re, obs_st)
        return obs, state, rewards, dones, infos

    @partial(jax.jit, static_argnums=[0])
    def unnormalise_obs(self, obs_batch: chex.Array) -> chex.Array:
        lidar = self.unnormalise_lidar(obs_batch[:, :self.lidar_num_beams])
        vel_obs = (obs_batch[:, self.lidar_num_beams:self.lidar_num_beams+2] + jnp.array([0.5, 0.0])) * jnp.array([self.max_v, self.max_w]) 
        goal_dist = (obs_batch[:, self.lidar_num_beams+2:self.lidar_num_beams+3] + 0.5) * self.lidar_max_range
        goal_orient = obs_batch[:, self.lidar_num_beams+3:self.lidar_num_beams+4] * jnp.pi
        rew_lambda = obs_batch[:, -1] + 0.5
        vel_goal = jnp.concatenate([vel_obs, goal_dist, goal_orient, rew_lambda[:, None]], axis=1)
        o = jnp.concatenate([lidar, vel_goal], axis=1)
        return o
    
    def get_monitored_metrics(self):
        return ["NumC", "GoalR", "AgentC", "MapC", "TimeO", "Return"]

    def get_eval_solved_rate_fn(self):
        def _fn(ep_stats):
            return ep_stats["GoalR"] / ep_stats["NumC"]
            
        return _fn

    def agent_action_space(self):
        if self._act_type == "Discrete":
            return Discrete(15)
        low = jnp.array(
            [self.min_v, -jnp.pi/6], dtype=jnp.float32 # NOTE hard coded heading angle
        )
        high = jnp.array(
            [self.max_v, jnp.pi/6], dtype=jnp.float32 
        )
        return Box(low, high, (2,), jnp.float32)
            
    
    def agent_observation_space(self):
        return Box(-jnp.inf, jnp.inf, (self.lidar_num_beams+5,))  # NOTE hardcoded
    
    # def action_space(self, agent=None):   # NOTE assuming homogenous observation spaces, NOTE I think jnp.empty is fine
    #     aa = self.agent_action_space()
    #     return jnp.empty((self.num_agents, *aa.shape))
    
    # def observation_space(self, agent=None):
    #     oo = self.agent_observation_space()
    #     return jnp.empty((self.num_agents, *oo.shape))
    
    @partial(jax.jit, static_argnums=[0])
    def generate_scenario(self, key):
        """ Sample map grid and agent start/goal poses """
        return self._map_obj.sample_scenario(key)
    
    def get_env_metrics(self, state: State) -> dict:
        """ NOTE only valid for grid map type"""
        # n_walls = state.map_data.sum() - state.map_data.shape[0]*2 - state.map_data.shape[1]*2 + 4
        inside = state.map_data.astype(jnp.bool_)[1:-1, 1:-1]
        n_walls = jnp.sum(inside)

        passable, path_len = jax.vmap(
            self.map_obj.dikstra_path,
            in_axes=(None, 0, 0)
        )(
            state.map_data,
            state.pos,
            state.goal,
        )
                
        shortest_path_lengths_stderr = jax.lax.select(
            jnp.sum(passable) > 0,
            jnp.std(path_len, where=passable)/jnp.sqrt(jnp.sum(passable)),
            0.0
        )
        return dict(
            n_walls=n_walls,
            shortest_path_length_mean=jnp.mean(path_len, where=passable),
            shortest_path_lengths_stderr=shortest_path_lengths_stderr,
            passable=jnp.mean(passable),
        )
        
    ### === VISULISATION === ###

    def plot_lidar(self, ax: axes.Axes, obs: Dict, state: State, num_to_plot: int=10):

        @partial(jax.vmap, in_axes=(0, 0, 0, None))
        def lidar_scatter(
            pos: chex.Array,
            theta: float,
            lidar_ranges: chex.Array,
            idx,
        ):
            """Return lidar ranges as points ready to be plotted with `ax.scatter()`

            Args:
                state (EnvSingleAState): agent state
                params (EnvParams): environment parameters
                ranges (chex.Array): reported lidar ranges

            Returns:
                Tuple[List, List]: lists of x and y coordinates respectively of lidar ranges for plotting
            """
            ranges = self.unnormalise_lidar(lidar_ranges) # (lidar_ranges+0.5)*self.lidar_max_range  # correct normalisation
            x = [ranges[i]*jnp.cos(self.lidar_angles[idx[i]]+theta) + pos[0] for i in range(ranges.shape[0])]
            y = [ranges[i]*jnp.sin(self.lidar_angles[idx[i]]+theta) + pos[1] for i in range(ranges.shape[0])]
            return jnp.array([x, y])  

        if self.lidar_num_beams>10:
            if num_to_plot > self.lidar_num_beams: 
                num_to_plot = self.lidar_num_beams
                print('Warning: num_to_plot > lidar_num_beams, setting num_to_plot to lidar_num_beams')
            idx = jnp.round(jnp.linspace(0, self.lidar_num_beams-1, num_to_plot)).astype(int)
        else:
            idx = range(self.lidar_num_beams)

        obs_batch = jnp.stack([obs[a] for a in self.agents])

        lidar_scat = lidar_scatter(state.pos, state.theta, obs_batch[:, idx], idx)
        lidar_scat = jnp.swapaxes(lidar_scat, 1, 2).reshape((-1, 2))
        lidar_scat = self._map_obj.scale_coords(lidar_scat)
        ax.scatter(lidar_scat[:, 0], lidar_scat[:, 1], c='b',  s=2)

    # Plotting by SMAX style
    def init_render(self, 
                    ax: axes.Axes, 
                    state: State, 
                    obs: Dict=None, 
                    lidar=True,  # plot lidar?
                    agent=True,  # plot agents?
                    goal=True,  # plot goals?
                    rew_lambda=False,  # plot lambda?
                    ticks_off=False,  # turn off axis ticks?
                    num_to_plot=10,  # number of lidar beams to plot
                    colour_agents_by_idx=False,
                    ):
        """ Render environment. """
 
        ax.set_aspect('equal', 'box')
        
        self.map_obj.plot_map(ax, state.map_data)
        if agent: 
            self.map_obj.plot_agents(ax, state.pos, state.theta, state.goal, state.done, plot_line_to_goal=goal, colour_agents_by_idx=colour_agents_by_idx)
        if lidar: 
            assert obs is not None, "lidar is True but no obs provided, TODO make it not obs dependent"
            self.plot_lidar(ax, obs, state, num_to_plot=num_to_plot)
        
        if rew_lambda:
            ax.text(0.5, 0.5, f"lambda: {state.rew_lambda}", fontsize=12, ha='center', va='center', c='white')
        if ticks_off:
            ax.set_xticks([])
            ax.set_yticks([])

        canvas = ax.figure.canvas
        canvas.draw()

    



import jax
import jax.numpy as jnp
import chex
# from .level import Level
# from .env import DIR_TO_VEC
from .jaxnav_env import State
from .maps import GridMapPolygonAgents
from enum import IntEnum
import numpy as np
from typing import Callable
from functools import partial



""" 
Mutation strategy:
1. flip wall: compute free cells, choose one at random, flip it to a wall
- should probably switch to a MapData class rather than just an array
  as it would be good to store free cells. Ignore for now

"""    
    

def make_level_mutator(max_num_edits: int, map: GridMapPolygonAgents):
    
    class Mutations(IntEnum):
        # Turn left, turn right, move forward
        NO_OP = 0
        FLIP_WALL = 1
        MOVE_GOAL = 2
        
    
    def move_goal_flip_walls(rng, state: State, n: int = 1):
        def _mutate(carry, step):
            state = carry
            rng, mutation = step

            def _apply(rng, state):    
                rng, arng, brng = jax.random.split(rng, 3)

                is_flip_wall = jnp.equal(mutation, Mutations.FLIP_WALL.value)
                mutated_state = flip_wall(arng, map, state)
                next_state = jax.tree.map(lambda x,y: jax.lax.select(is_flip_wall, x, y), mutated_state, state)

                is_move_goal = jnp.equal(mutation, Mutations.MOVE_GOAL.value)
                mutated_state = move_goal(brng, map, state)
                next_state = jax.tree.map(lambda x,y: jax.lax.select(is_move_goal, x, y), mutated_state, next_state)
                
                return next_state
                
            return jax.lax.cond(mutation != -1, _apply, lambda *_: state, rng, state), None
        
        
        rng, nrng, *mrngs = jax.random.split(rng, max_num_edits+2)
        mutations = jax.random.choice(nrng, np.arange(len(Mutations)), (max_num_edits,))
        mutations = jnp.where(jnp.arange(max_num_edits) < n, mutations, -1) # mask out extra mutations

        new_level, _ = jax.lax.scan(_mutate, state, (jnp.array(mrngs), mutations))

        return new_level
    
    return move_goal_flip_walls
    
def flip_wall(rng, map: GridMapPolygonAgents, state: State):
    wall_map = state.map_data
    h,w = wall_map.shape
        
    # goal_map_mask = jnp.any(jax.vmap(
    #     map.get_circle_map_occupancy_mask,
    #     in_axes=(0, None, None)
    # )(state.goal, wall_map, 0.3), axis=0)
    
    goal_map_mask = wall_map

    start_map_mask = jnp.any(jax.vmap(
        map.get_agent_map_occupancy_mask, in_axes=(0,0,None)
    )(state.pos, state.theta, wall_map), axis=0)
    

    goal = jnp.floor(state.goal).astype(jnp.int32)
    goal_map_mask = goal_map_mask.at[goal[:, 1], goal[:, 0]].set(1)

    map_mask = start_map_mask | map._gen_base_grid() | goal_map_mask
    

    flip_idx = jax.random.choice(rng, np.arange(h*w), p=jnp.logical_not(map_mask.flatten()))

    flip_y = flip_idx // w
    flip_x = flip_idx % w
    
    flip_val = 1-wall_map.at[flip_y, flip_x].get()
    next_wall_map = wall_map.at[flip_y, flip_x].set(flip_val)
    print('next_wall_map', next_wall_map)
    return state.replace(map_data=next_wall_map)


def move_goal(rng, map:GridMapPolygonAgents, state:State):
    wall_map = state.map_data
    h,w = wall_map.shape
    
    rng, _rng = jax.random.split(rng)
    agent_idx = jax.random.choice(_rng, np.arange(state.pos.shape[0]))
    
    # goal_map_masks = jax.vmap(
    #     map.get_circle_map_occupancy_mask,
    #     in_axes=(0, None, None)
    # )(state.goal, wall_map, 1.0)
    goal_map_masks = jnp.repeat(wall_map[None], state.pos.shape[0], axis=0)
    # goal_map_masks = goal_map_masks.at[agent_idx].set(jnp.zeros(wall_map.shape, dtype=jnp.int32))
    goal_map_mask = jnp.any(goal_map_masks, axis=0)
    current_goal = jnp.floor(state.goal[agent_idx]).astype(jnp.int32)
    goal_map_mask = goal_map_mask.at[current_goal[1], current_goal[0]].set(0)
    
    start_map_mask = jnp.any(jax.vmap(
        map.get_agent_map_occupancy_mask, in_axes=(0,0,None)
    )(state.pos, state.theta, wall_map), axis=0)
    
    map_mask = start_map_mask | map._gen_base_grid() | goal_map_mask
    
    next_idx = jax.random.choice(rng, np.arange(h*w), p=jnp.logical_not(map_mask.flatten()))
    next_goal_y = next_idx // w
    next_goal_x = next_idx %  w
    
    goals = state.goal.at[agent_idx].set(jnp.array([next_goal_x, next_goal_y]) + 0.5) # Make the goal in the center of the cell.
    
    return state.replace(goal=goals)
    
    
    
    


from PIL import Image
import os
import jax
import jax.numpy as jnp

from jaxmarl import make
from jaxmarl.environments.storm.storm_2p import Items

action = 1
render_agent_view = True
num_outer_steps = 3
num_inner_steps = 152

rng = jax.random.PRNGKey(0)

env = make('storm_2p', 
        num_inner_steps=num_inner_steps, 
        num_outer_steps=num_outer_steps, 
        fixed_coin_location=True,
        num_agents=2,
        payoff_matrix=jnp.array([[[3, 0], [5, 1]], [[3, 5], [0, 1]]]),
        freeze_penalty=5,)

num_actions = env.action_space().n

obs, old_state = env.reset(rng)
pics = []
pics1 = []
pics2 = []

img = env.render(old_state)
img1 = env.render_agent_view(old_state, agent=0)
img2 = env.render_agent_view(old_state, agent=1)
pics.append(img)

int_action = {
    0: "left",
    1: "right",
    2: "forward",
    3: "interact",
    4: "stay",
}

key_int = {"w": 2, "a": 0, "s": 4, "d": 1, " ": 4}
env.step_env = jax.jit(env.step_env)

for t in range(num_outer_steps * num_inner_steps):
    rng, rng1, rng2 = jax.random.split(rng, 3)
    # a1 = jnp.array(2)
    # a2 = jnp.array(4)
    a1 = jax.random.choice(
        rng1, a=num_actions, p=jnp.array([0.1, 0.1, 0.5, 0.1, 0.4])
    )
    a2 = jax.random.choice(
        rng2, a=num_actions, p=jnp.array([0.1, 0.1, 0.5, 0.1, 0.4])
    )
    obs, state, reward, done, info = env.step_env(
        rng, old_state, (a1 * action, a2 * action)
    )

    print('outer t', state.outer_t)
    print('inner t', state.inner_t)
    print('done', done)
    if (state.red_pos[:2] == state.blue_pos[:2]).all():
        import pdb

        # pdb.set_trace()
        print("collision")
        print(
            f"timestep: {t}, A1: {int_action[a1.item()]} A2:{int_action[a2.item()]}"
        )
        print(state.red_pos, state.blue_pos)

    img = env.render(state)
    pics.append(img)

    if render_agent_view:
        img1 = env.render_agent_view(state, agent=0)
        img2 = env.render_agent_view(state, agent=1)
        pics1.append(img1)
        pics2.append(img2)
    old_state = state
print("Saving GIF")
pics = [Image.fromarray(img) for img in pics]
pics[0].save(
    "state.gif",
    format="GIF",
    save_all=True,
    optimize=False,
    append_images=pics[1:],
    duration=100,
    loop=0,
)

if render_agent_view:
    pics1 = [Image.fromarray(img) for img in pics1]
    pics2 = [Image.fromarray(img) for img in pics2]
    pics1[0].save(
        "agent1.gif",
        format="GIF",
        save_all=True,
        optimize=False,
        append_images=pics1[1:],
        duration=100,
        loop=0,
    )
    pics2[0].save(
        "agent2.gif",
        format="GIF",
        save_all=True,
        optimize=False,
        append_images=pics2[1:],
        duration=100,
        loop=0,
    )


URGENCY_CUTOFF = 40  # When this many time steps remain, the urgency layer is flipped on
DELIVERY_REWARD = 20
POT_COOK_TIME = 20  # Time it takes to cook a pot
INDICATOR_ACTIVATION_TIME = 10  # Time the recipe indicator is active for
INDICATOR_ACTIVATION_COST = 5  # Cost of activating the recipe indicator

SHAPED_REWARDS = {
    "PLACEMENT_IN_POT": 3,
    "POT_START_COOKING": 5,
    # "DISH_PICKUP": 0,
    # "PLATE_PICKUP": 0,
    "DISH_PICKUP": 5,
    "PLATE_PICKUP": 3,
}
